"""Sequence management and execution."""

import asyncio
import json
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Optional

from pydantic import BaseModel, Field

from scopinator.sequencer.base import Command, CommandStatus
from scopinator.sequencer.commands import (
    GoToTargetCommand,
    SequenceCommand,
    StartImagingCommand,
    StopImagingCommand,
    WaitMinutesCommand,
    WaitUntilEventCommand,
    WaitUntilTimeCommand,
)


class SequenceState(str, Enum):
    """State of sequence execution."""

    IDLE = "idle"
    RUNNING = "running"
    PAUSED = "paused"
    COMPLETED = "completed"
    FAILED = "failed"
    CANCELLED = "cancelled"


class Sequence(BaseModel):
    """A sequence of commands for telescope and imaging automation.

    Sequences can be saved to and loaded from JSON files.
    """

    name: str = Field(..., description="Name of the sequence")
    description: Optional[str] = Field(None, description="Description of the sequence")
    commands: list[Command] = Field(default_factory=list, description="List of commands to execute")
    state: SequenceState = Field(default=SequenceState.IDLE, description="Current sequence state")
    created_at: datetime = Field(default_factory=datetime.utcnow, description="When sequence was created")
    started_at: Optional[datetime] = Field(None, description="When sequence started")
    completed_at: Optional[datetime] = Field(None, description="When sequence completed")
    current_command_index: int = Field(0, description="Index of currently executing command")
    _execution_task: Optional[asyncio.Task] = None

    class Config:
        """Pydantic configuration."""

        arbitrary_types_allowed = True
        use_enum_values = False

    async def start(self, context: dict[str, Any]) -> None:
        """Start executing the sequence.

        Args:
            context: Execution context containing telescope client, state, etc.

        Raises:
            RuntimeError: If sequence is already running
        """
        if self.state == SequenceState.RUNNING:
            raise RuntimeError("Sequence is already running")

        self.state = SequenceState.RUNNING
        self.started_at = datetime.utcnow()
        self._execution_task = asyncio.create_task(self._execute(context))

    async def stop(self) -> None:
        """Stop the sequence execution."""
        if self.state == SequenceState.RUNNING:
            self.state = SequenceState.CANCELLED

            # Cancel currently running command
            if 0 <= self.current_command_index < len(self.commands):
                current_cmd = self.commands[self.current_command_index]
                if current_cmd.status == CommandStatus.RUNNING:
                    await current_cmd.cancel()

            if self._execution_task and not self._execution_task.done():
                self._execution_task.cancel()
                try:
                    await self._execution_task
                except asyncio.CancelledError:
                    pass

            self.completed_at = datetime.utcnow()

    async def pause(self) -> None:
        """Pause the sequence execution."""
        if self.state == SequenceState.RUNNING:
            self.state = SequenceState.PAUSED

            # Pause currently running command
            if 0 <= self.current_command_index < len(self.commands):
                current_cmd = self.commands[self.current_command_index]
                if current_cmd.status == CommandStatus.RUNNING:
                    await current_cmd.pause()

    async def resume(self, context: dict[str, Any]) -> None:
        """Resume the sequence execution.

        Args:
            context: Execution context

        Raises:
            RuntimeError: If sequence is not paused
        """
        if self.state != SequenceState.PAUSED:
            raise RuntimeError("Sequence is not paused")

        self.state = SequenceState.RUNNING

        # Resume currently paused command
        if 0 <= self.current_command_index < len(self.commands):
            current_cmd = self.commands[self.current_command_index]
            if current_cmd.status == CommandStatus.PAUSED:
                await current_cmd.resume()

        # Restart execution task
        self._execution_task = asyncio.create_task(self._execute(context))

    async def _execute(self, context: dict[str, Any]) -> None:
        """Internal execution loop.

        Args:
            context: Execution context
        """
        try:
            while self.current_command_index < len(self.commands):
                if self.state == SequenceState.CANCELLED:
                    break

                if self.state == SequenceState.PAUSED:
                    # Wait until resumed
                    while self.state == SequenceState.PAUSED:
                        await asyncio.sleep(0.1)

                cmd = self.commands[self.current_command_index]

                # Skip already completed commands (in case of resume)
                if cmd.status == CommandStatus.COMPLETED:
                    self.current_command_index += 1
                    continue

                try:
                    await cmd.execute(context)

                    if cmd.status == CommandStatus.FAILED:
                        self.state = SequenceState.FAILED
                        self.completed_at = datetime.utcnow()
                        return

                except asyncio.CancelledError:
                    self.state = SequenceState.CANCELLED
                    raise
                except Exception as e:
                    cmd.mark_failed(str(e))
                    self.state = SequenceState.FAILED
                    self.completed_at = datetime.utcnow()
                    raise

                self.current_command_index += 1

            if self.state == SequenceState.RUNNING:
                self.state = SequenceState.COMPLETED
                self.completed_at = datetime.utcnow()

        except asyncio.CancelledError:
            self.state = SequenceState.CANCELLED
            self.completed_at = datetime.utcnow()

    def to_json(self, file_path: Optional[str] = None) -> str:
        """Serialize sequence to JSON.

        Args:
            file_path: Optional path to save JSON file

        Returns:
            str: JSON representation of the sequence
        """
        # Use Pydantic's model_dump with custom serialization
        data = self.model_dump(mode="json", exclude={"_execution_task"})

        # Convert to JSON string
        json_str = json.dumps(data, indent=2, default=str)

        # Optionally save to file
        if file_path:
            Path(file_path).write_text(json_str)

        return json_str

    @classmethod
    def from_json(cls, json_str: Optional[str] = None, file_path: Optional[str] = None) -> "Sequence":
        """Deserialize sequence from JSON.

        Args:
            json_str: JSON string to deserialize
            file_path: Optional path to load JSON file from

        Returns:
            Sequence: Deserialized sequence

        Raises:
            ValueError: If neither json_str nor file_path is provided
        """
        if file_path:
            json_str = Path(file_path).read_text()
        elif json_str is None:
            raise ValueError("Either json_str or file_path must be provided")

        data = json.loads(json_str)

        # Reconstruct command objects from discriminated union
        commands = []
        for cmd_data in data.get("commands", []):
            cmd_type = cmd_data.get("name")  # Use name field to determine type
            # You could also add a 'type' field for more explicit typing

            # Map command data to appropriate class
            # This is a simple approach - for production use discriminated unions
            commands.append(_deserialize_command(cmd_data))

        data["commands"] = commands
        return cls(**data)


def _deserialize_command(cmd_data: dict[str, Any]) -> Command:
    """Deserialize a command from dictionary.

    Args:
        cmd_data: Command data dictionary

    Returns:
        Command: Deserialized command instance
    """
    # Map of command types - in production, use Pydantic discriminated unions
    command_classes = {
        "WaitMinutesCommand": WaitMinutesCommand,
        "WaitUntilTimeCommand": WaitUntilTimeCommand,
        "WaitUntilEventCommand": WaitUntilEventCommand,
        "GoToTargetCommand": GoToTargetCommand,
        "StartImagingCommand": StartImagingCommand,
        "StopImagingCommand": StopImagingCommand,
        "SequenceCommand": SequenceCommand,
    }

    # Determine command type from the data
    # Try to infer from field presence or use explicit 'command_type' field
    cmd_type = cmd_data.get("command_type")

    if not cmd_type:
        # Infer type from fields
        if "minutes" in cmd_data:
            cmd_type = "WaitMinutesCommand"
        elif "target_time" in cmd_data:
            cmd_type = "WaitUntilTimeCommand"
        elif "event" in cmd_data:
            cmd_type = "WaitUntilEventCommand"
        elif "ra" in cmd_data and "dec" in cmd_data:
            cmd_type = "GoToTargetCommand"
        elif "exposure_time" in cmd_data:
            cmd_type = "StartImagingCommand"
        elif "commands" in cmd_data:
            cmd_type = "SequenceCommand"
        else:
            cmd_type = "StopImagingCommand"

    command_class = command_classes.get(cmd_type)
    if not command_class:
        raise ValueError(f"Unknown command type: {cmd_type}")

    # Handle nested commands for SequenceCommand
    if cmd_type == "SequenceCommand":
        nested_commands = [_deserialize_command(c) for c in cmd_data.get("commands", [])]
        cmd_data["commands"] = nested_commands

    return command_class(**cmd_data)



"""Seestar backend - adapter for existing SeestarClient."""

from scopinator.v2.backends.seestar.backend import SeestarBackend
from scopinator.v2.backends.seestar.mount import SeestarMount
from scopinator.v2.backends.seestar.camera import SeestarCamera

__all__ = [
    "SeestarBackend",
    "SeestarMount",
    "SeestarCamera",
]



"""Seestar camera implementation.

Wraps the existing SeestarClient and SeestarImagingClient to provide
the V2 Camera interface.
"""

from typing import TYPE_CHECKING, AsyncIterator, Optional

from scopinator.v2.core.capabilities import CameraCapabilities
from scopinator.v2.core.devices import Camera, CameraStatus
from scopinator.v2.core.events import EventType, ExposureEvent, UnifiedEventBus
from scopinator.v2.core.exceptions import DeviceError, NotConnectedError
from scopinator.v2.core.types import CameraState, ExposureSettings, ImageData

if TYPE_CHECKING:
    from scopinator.seestar.client import SeestarClient
    from scopinator.seestar.imaging_client import SeestarImagingClient


class SeestarCamera(Camera):
    """Camera implementation for Seestar.

    This adapter wraps both the SeestarClient (for control commands)
    and SeestarImagingClient (for image data) to provide the V2
    Camera interface.

    Note: Seestar operates primarily in stacking mode, where the
    telescope continuously captures and stacks frames. This adapter
    maps that workflow to the Camera interface.
    """

    def __init__(
        self,
        client: "SeestarClient",
        imaging_client: Optional["SeestarImagingClient"] = None,
        event_bus: Optional[UnifiedEventBus] = None,
    ) -> None:
        """Initialize Seestar camera adapter.

        Args:
            client: Main SeestarClient for control
            imaging_client: Optional imaging client for image data
            event_bus: Optional event bus for emitting camera events
        """
        super().__init__(event_bus)
        self._client = client
        self._imaging_client = imaging_client
        self._current_settings: Optional[ExposureSettings] = None

    async def connect(self) -> None:
        """Camera connection is handled by backend/client."""
        self._connected = self._client.is_connected

    async def disconnect(self) -> None:
        """Camera disconnection is handled by backend/client."""
        pass

    async def get_capabilities(self) -> CameraCapabilities:
        """Get Seestar camera capabilities."""
        return CameraCapabilities(
            can_abort_exposure=True,
            can_stop_exposure=False,
            can_set_gain=True,
            can_set_offset=False,
            can_set_binning=False,  # Seestar has fixed binning
            can_subframe=False,
            can_fast_readout=False,
            has_shutter=False,
            has_cooler=False,
            has_filter_wheel=True,  # LP filter
            is_color=True,  # Bayer sensor
            max_bin_x=1,
            max_bin_y=1,
            sensor_width=1920,  # Seestar S50 resolution
            sensor_height=1080,
            pixel_size_x=3.0,  # Approximate
            pixel_size_y=3.0,
            min_exposure=0.01,
            max_exposure=20.0,  # Seestar typical max
            supported_gains=[],  # Gain is managed automatically
            bit_depth=12,
        )

    async def start_exposure(self, settings: ExposureSettings) -> None:
        """Start an exposure/stacking session.

        For Seestar, this starts the stacking mode with the given
        exposure settings.

        Args:
            settings: Exposure settings
        """
        if not self._client.is_connected:
            raise NotConnectedError("Seestar is not connected")

        self._current_settings = settings

        # Emit exposure started event
        event = ExposureEvent(
            event_type=EventType.EXPOSURE_STARTED,
            source_device="seestar_camera",
            source_backend="seestar",
            duration_seconds=settings.duration_seconds,
        )
        self._event_bus.emit_nowait(event)

        # For Seestar, stacking is typically started via goto
        # or scope_view command. The exposure settings are applied
        # automatically by the telescope.

        # If we have a current view, we're already exposing
        # Otherwise, this is typically called after a goto

    async def abort_exposure(self) -> None:
        """Abort current exposure/stacking."""
        if not self._client.is_connected:
            raise NotConnectedError("Seestar is not connected")

        await self._client.stop_stack()

        event = ExposureEvent(
            event_type=EventType.EXPOSURE_ABORTED,
            source_device="seestar_camera",
            source_backend="seestar",
        )
        self._event_bus.emit_nowait(event)

    async def get_image(self) -> ImageData:
        """Get the last captured/stacked image.

        Returns:
            ImageData containing the stacked image

        Raises:
            DeviceError: If no image available
        """
        if self._imaging_client is None:
            raise DeviceError("Imaging client not configured")

        if not self._imaging_client.is_connected:
            raise NotConnectedError("Imaging client is not connected")

        # Get the last image from the imaging client
        scope_image = await self._imaging_client.get_last_image()

        if scope_image is None or scope_image.data is None:
            raise DeviceError("No image available")

        return ImageData(
            width=scope_image.width or 1920,
            height=scope_image.height or 1080,
            data=scope_image.data,
            bit_depth=12,
            is_color=True,
            bayer_pattern="GRBG",
            metadata={
                "stacked_frames": self._client.status.stacked_frame,
                "dropped_frames": self._client.status.dropped_frame,
                "target_name": self._client.status.target_name,
            },
        )

    async def is_exposing(self) -> bool:
        """Check if camera is currently exposing/stacking."""
        return self._client.client_mode in ("Stack", "ContinuousExposure")

    async def stream_images(
        self, settings: ExposureSettings
    ) -> AsyncIterator[ImageData]:
        """Stream images from the camera.

        Yields stacked images as they become available.

        Args:
            settings: Exposure settings (not all may be applicable)

        Yields:
            ImageData for each captured/stacked frame
        """
        if self._imaging_client is None:
            raise DeviceError("Imaging client not configured")

        if not self._imaging_client.is_connected:
            raise NotConnectedError("Imaging client is not connected")

        # Use the imaging client's image generator
        async for scope_image in self._imaging_client.get_next_image():
            if scope_image.data is not None:
                yield ImageData(
                    width=scope_image.width or 1920,
                    height=scope_image.height or 1080,
                    data=scope_image.data,
                    bit_depth=12,
                    is_color=True,
                    bayer_pattern="GRBG",
                    metadata={
                        "stacked_frames": self._client.status.stacked_frame,
                        "dropped_frames": self._client.status.dropped_frame,
                    },
                )

    async def set_gain(self, gain: int) -> None:
        """Set camera gain.

        Seestar manages gain automatically, but this can override.
        """
        # Seestar has automatic gain control
        # Could implement via SetSetting command if needed
        pass

    async def get_status(self) -> CameraStatus:
        """Get current camera status."""
        client_mode = self._client.client_mode

        # Map client mode to camera state
        if client_mode == "Stack":
            state = CameraState.EXPOSING
        elif client_mode == "ContinuousExposure":
            state = CameraState.EXPOSING
        elif client_mode == "Idle":
            state = CameraState.IDLE
        else:
            state = CameraState.IDLE

        return CameraStatus(
            connected=self._client.is_connected,
            name="Seestar Camera",
            driver_info="Seestar V2 Adapter",
            driver_version="1.0.0",
            state=state,
            exposure_progress=0.0,  # Could calculate from frame count
            cooler_on=False,
            cooler_setpoint=None,
            temperature=self._client.status.temp,
            cooler_power=None,
            gain=self._client.status.gain,
            offset=None,
            binning=(1, 1),
            last_exposure_duration=None,
        )



"""Device manager for orchestrating multiple backends.

The DeviceManager provides a unified interface for working with devices
across multiple protocol backends (Seestar, Alpaca, INDI). It handles
backend lifecycle, device caching, and event bus distribution.
"""

from typing import Optional

from scopinator.v2.backends.base import Backend
from scopinator.v2.core.devices import Camera, FilterWheel, Focuser, Mount
from scopinator.v2.core.events import UnifiedEventBus


class DeviceManager:
    """Manages multiple backends and provides unified device access.

    The DeviceManager is the main entry point for multi-protocol telescope
    control. It allows registering multiple backends and provides a unified
    way to access devices regardless of their underlying protocol.

    Example:
        manager = DeviceManager()

        # Add backends
        await manager.add_backend("seestar", SeestarBackend("192.168.1.100"))
        await manager.add_backend("alpaca", AlpacaBackend("localhost"))

        # Discover devices from all backends
        all_devices = await manager.discover_all()
        # {
        #     "seestar": {"mount": ["seestar_mount"], "camera": ["seestar_camera"]},
        #     "alpaca": {"mount": ["mount_0"], "camera": ["camera_0"]}
        # }

        # Get a specific device
        mount = await manager.get_mount("seestar", "seestar_mount")
        await mount.slew_to_coordinates(Coordinates(ra=83.63, dec=22.01))
    """

    def __init__(self, event_bus: Optional[UnifiedEventBus] = None) -> None:
        """Initialize device manager.

        Args:
            event_bus: Optional shared event bus. If not provided, a new
                      one will be created and shared with all backends.
        """
        self._event_bus = event_bus or UnifiedEventBus()
        self._backends: dict[str, Backend] = {}
        self._mounts: dict[str, Mount] = {}
        self._cameras: dict[str, Camera] = {}
        self._focusers: dict[str, Focuser] = {}
        self._filterwheels: dict[str, FilterWheel] = {}

    @property
    def event_bus(self) -> UnifiedEventBus:
        """Get the shared event bus."""
        return self._event_bus

    @property
    def backends(self) -> dict[str, Backend]:
        """Get all registered backends."""
        return dict(self._backends)

    def get_backend(self, name: str) -> Optional[Backend]:
        """Get a backend by name.

        Args:
            name: Backend name

        Returns:
            Backend instance or None if not found
        """
        return self._backends.get(name)

    async def add_backend(self, name: str, backend: Backend) -> None:
        """Add and connect a backend.

        The backend will share the manager's event bus for unified
        event handling.

        Args:
            name: Unique name for this backend
            backend: Backend instance to add

        Raises:
            ValueError: If a backend with this name already exists
        """
        if name in self._backends:
            raise ValueError(f"Backend '{name}' already exists")

        # Share the event bus
        backend.event_bus = self._event_bus

        # Connect the backend
        await backend.connect()

        self._backends[name] = backend

    async def remove_backend(self, name: str) -> None:
        """Disconnect and remove a backend.

        Args:
            name: Backend name to remove
        """
        if name in self._backends:
            backend = self._backends[name]
            await backend.disconnect()
            del self._backends[name]

            # Remove cached devices from this backend
            prefix = f"{name}:"
            self._mounts = {k: v for k, v in self._mounts.items() if not k.startswith(prefix)}
            self._cameras = {k: v for k, v in self._cameras.items() if not k.startswith(prefix)}
            self._focusers = {k: v for k, v in self._focusers.items() if not k.startswith(prefix)}
            self._filterwheels = {
                k: v for k, v in self._filterwheels.items() if not k.startswith(prefix)
            }

    async def discover_all(self) -> dict[str, dict[str, list[str]]]:
        """Discover devices from all backends.

        Returns:
            Dict mapping backend name to device discovery results.

            Example:
                {
                    "seestar": {
                        "mount": ["seestar_mount"],
                        "camera": ["seestar_camera"]
                    },
                    "alpaca": {
                        "mount": ["mount_0"],
                        "camera": ["camera_0"]
                    }
                }
        """
        result: dict[str, dict[str, list[str]]] = {}
        for name, backend in self._backends.items():
            try:
                result[name] = await backend.discover_devices()
            except Exception:
                result[name] = {}
        return result

    async def get_mount(self, backend_name: str, device_id: str) -> Mount:
        """Get a mount from a specific backend.

        Devices are cached, so calling this multiple times with the same
        arguments returns the same instance.

        Args:
            backend_name: Name of the backend
            device_id: Device ID within that backend

        Returns:
            Mount instance

        Raises:
            ValueError: If backend not found
            DeviceError: If device not found
        """
        key = f"{backend_name}:{device_id}"

        if key not in self._mounts:
            backend = self._backends.get(backend_name)
            if not backend:
                raise ValueError(f"Unknown backend: {backend_name}")

            mount = await backend.get_mount(device_id)
            self._mounts[key] = mount

        return self._mounts[key]

    async def get_camera(self, backend_name: str, device_id: str) -> Camera:
        """Get a camera from a specific backend.

        Args:
            backend_name: Name of the backend
            device_id: Device ID within that backend

        Returns:
            Camera instance

        Raises:
            ValueError: If backend not found
            DeviceError: If device not found
        """
        key = f"{backend_name}:{device_id}"

        if key not in self._cameras:
            backend = self._backends.get(backend_name)
            if not backend:
                raise ValueError(f"Unknown backend: {backend_name}")

            camera = await backend.get_camera(device_id)
            self._cameras[key] = camera

        return self._cameras[key]

    async def get_focuser(self, backend_name: str, device_id: str) -> Focuser:
        """Get a focuser from a specific backend.

        Args:
            backend_name: Name of the backend
            device_id: Device ID within that backend

        Returns:
            Focuser instance

        Raises:
            ValueError: If backend not found
            DeviceError: If device not found
        """
        key = f"{backend_name}:{device_id}"

        if key not in self._focusers:
            backend = self._backends.get(backend_name)
            if not backend:
                raise ValueError(f"Unknown backend: {backend_name}")

            focuser = await backend.get_focuser(device_id)
            self._focusers[key] = focuser

        return self._focusers[key]

    async def get_filterwheel(self, backend_name: str, device_id: str) -> FilterWheel:
        """Get a filter wheel from a specific backend.

        Args:
            backend_name: Name of the backend
            device_id: Device ID within that backend

        Returns:
            FilterWheel instance

        Raises:
            ValueError: If backend not found
            DeviceError: If device not found
        """
        key = f"{backend_name}:{device_id}"

        if key not in self._filterwheels:
            backend = self._backends.get(backend_name)
            if not backend:
                raise ValueError(f"Unknown backend: {backend_name}")

            filterwheel = await backend.get_filterwheel(device_id)
            self._filterwheels[key] = filterwheel

        return self._filterwheels[key]

    async def disconnect_all(self) -> None:
        """Disconnect all backends and clear cached devices."""
        for name in list(self._backends.keys()):
            await self.remove_backend(name)

    async def __aenter__(self) -> "DeviceManager":
        """Async context manager entry."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> bool:
        """Async context manager exit - disconnects all backends."""
        await self.disconnect_all()
        return False

    def __repr__(self) -> str:
        backend_names = list(self._backends.keys())
        return f"DeviceManager(backends={backend_names})"



"""Connection state persistence for CLI.

Saves and loads connection state between CLI invocations.
"""

import json
from pathlib import Path
from typing import Any, Optional

# State file location
SCOPINATOR_DIR = Path.home() / ".scopinator"
CONNECTION_STATE_FILE = SCOPINATOR_DIR / "connection.json"


def ensure_state_dir() -> Path:
    """Ensure the state directory exists."""
    SCOPINATOR_DIR.mkdir(parents=True, exist_ok=True)
    return SCOPINATOR_DIR


def save_connection_state(
    protocol: str,
    host: str,
    port: int,
    extra: Optional[dict[str, Any]] = None,
) -> None:
    """Save connection state to disk.

    Args:
        protocol: Protocol name (seestar, alpaca, indi)
        host: Telescope host/IP
        port: Telescope port
        extra: Optional extra data to save
    """
    ensure_state_dir()

    state = {
        "protocol": protocol,
        "host": host,
        "port": port,
    }
    if extra:
        state.update(extra)

    with open(CONNECTION_STATE_FILE, "w") as f:
        json.dump(state, f, indent=2)


def load_connection_state() -> Optional[dict[str, Any]]:
    """Load connection state from disk.

    Returns:
        Connection state dict or None if no state saved
    """
    if not CONNECTION_STATE_FILE.exists():
        return None

    try:
        with open(CONNECTION_STATE_FILE, "r") as f:
            return json.load(f)
    except (json.JSONDecodeError, IOError):
        return None


def clear_connection_state() -> None:
    """Clear saved connection state."""
    if CONNECTION_STATE_FILE.exists():
        CONNECTION_STATE_FILE.unlink()


def get_saved_host() -> Optional[str]:
    """Get saved host from connection state."""
    state = load_connection_state()
    return state.get("host") if state else None


def get_saved_port() -> Optional[int]:
    """Get saved port from connection state."""
    state = load_connection_state()
    return state.get("port") if state else None


def get_saved_protocol() -> Optional[str]:
    """Get saved protocol from connection state."""
    state = load_connection_state()
    return state.get("protocol") if state else None



"""OpenCV Backend RTSP Client"""

import asyncio
import os
from threading import Thread, RLock
from typing import Any, Optional

import cv2
import numpy.typing as npt
from PIL import Image
from scopinator.util.logging_config import get_logger
logging = get_logger(__name__)

# Suppress FFmpeg/H.264 decoder warnings
# These are common with RTSP streams and usually harmless
os.environ["OPENCV_FFMPEG_CAPTURE_OPTIONS"] = "loglevel;quiet"
os.environ["OPENCV_LOG_LEVEL"] = "ERROR"  # Only show errors, not warnings


# Adapted from https://github.com/dactylroot/rtsp/tree/master


class RtspClient:
    """Maintain a live RTSP feed without buffering."""

    def __init__(self, rtsp_server_uri: str):
        self.rtsp_server_uri: str = rtsp_server_uri
        self._stream: cv2.VideoCapture | None = None
        self._bg_run: bool = False
        self._bgt: Thread | None = None
        self._queue: Image.Image | None = None
        self._lock = RLock()
        self._is_opening: bool = False

    def __enter__(self, *args, **kwargs):
        """Returns the object which later will have __exit__ called.
        This relationship creates a context manager."""
        self._bgt = Thread(target=self._opener, args=(), daemon=True)
        self._bgt.start()

        return self

    def __exit__(self, type=None, value=None, traceback=None):
        """Together with __enter__, allows support for `with-` clauses."""
        Thread(target=self._closer, args=(), daemon=True).start()

    def _opener(self):
        """open stream and start update process."""
        self._is_opening = True
        with self._lock:
            self._stream = cv2.VideoCapture(self.rtsp_server_uri)
            logging.debug("Connected to video source {}.".format(self.rtsp_server_uri))

            self._bg_run = True
        self._is_opening = False

        self._update()

    def _closer(self) -> None:
        """signal background thread to stop. release CV stream"""
        with self._lock:
            self._is_opening = False
            self._bg_run = False
            self._bgt.join()
            logging.debug("Disconnected from {}".format(self.rtsp_server_uri))

    def is_opened(self) -> bool:
        """return true if stream is opened and being read, else ensure closed"""
        try:
            return (
                (self._stream is not None) and self._stream.isOpened() and self._bg_run
            )
        except:
            self._close()
            return False

    def _update(self) -> None:
        while self.is_opened():
            (grabbed, frame) = self._stream.read()
            if not grabbed:
                self._bg_run = False
            else:
                with self._lock:
                    self._queue = frame

        self._stream.release()

    def read(self) -> Optional[npt.NDArray]:
        """Retrieve most recent frame and convert to PIL. Return unconverted with raw=True."""
        with self._lock:
            try:
                if self._queue is None:
                    return None
                return cv2.cvtColor(self._queue, cv2.COLOR_BGR2RGB)
            except:
                return None

    async def finish_opening(self):
        """Wait until stream is opened."""
        while True:
            with self._lock:
                if not self._is_opening:
                    logging.info("Stream opened.")
                    return
            await asyncio.sleep(0.1)


# def preview(self):
#     """ Blocking function. Opens OpenCV window to display stream. """
#     win_name = 'RTSP'
#     cv2.namedWindow(win_name, cv2.WINDOW_AUTOSIZE)
#     cv2.moveWindow(win_name,20,20)
#     while(self.isOpened()):
#         cv2.imshow(win_name,self.read(raw=True))
#         if cv2.waitKey(30) == ord('q'): # wait 30 ms for 'q' input
#             break
#     cv2.waitKey(1)
#     cv2.destroyAllWindows()
#     cv2.waitKey(1)



"""Example sequences for demonstration purposes."""

from datetime import datetime, timedelta

from scopinator.sequencer import (
    GoToTargetCommand,
    Sequence,
    SequenceCommand,
    StartImagingCommand,
    StopImagingCommand,
    WaitMinutesCommand,
    WaitUntilEventCommand,
    WaitUntilTimeCommand,
)
from scopinator.sequencer.events import AstronomicalEvent


def create_simple_imaging_sequence() -> Sequence:
    """Create a simple imaging sequence.

    This sequence:
    1. Waits 5 minutes
    2. Goes to a target (M31)
    3. Starts imaging
    4. Waits 60 minutes
    5. Stops imaging

    Returns:
        Sequence: A configured sequence
    """
    return Sequence(
        name="Simple M31 Imaging",
        description="Basic imaging sequence for M31 (Andromeda Galaxy)",
        commands=[
            WaitMinutesCommand(name="Initial wait", minutes=5),
            GoToTargetCommand(
                name="Slew to M31",
                ra=10.6847,  # M31 RA in degrees
                dec=41.2689,  # M31 Dec in degrees
                target_name="M31 (Andromeda Galaxy)",
            ),
            StartImagingCommand(
                name="Start imaging",
                exposure_time=300.0,  # 5 minute exposures
                gain=100,
                count=12,  # 12 exposures = 1 hour
            ),
            WaitMinutesCommand(name="Wait for imaging", minutes=60),
            StopImagingCommand(name="Stop imaging"),
        ],
    )


def create_dawn_sequence() -> Sequence:
    """Create a sequence that runs until astronomical dawn.

    This sequence:
    1. Waits until sunset
    2. Goes to target
    3. Starts imaging
    4. Waits until astronomical dawn
    5. Stops imaging

    Returns:
        Sequence: A configured sequence
    """
    # Example coordinates (San Francisco)
    latitude = 37.7749
    longitude = -122.4194

    return Sequence(
        name="All Night Imaging",
        description="Imaging sequence that runs from sunset to dawn",
        commands=[
            WaitUntilEventCommand(
                name="Wait for sunset",
                event=AstronomicalEvent.SUNSET,
                latitude=latitude,
                longitude=longitude,
            ),
            GoToTargetCommand(
                name="Slew to target",
                ra=83.8221,  # M42 RA
                dec=-5.3911,  # M42 Dec
                target_name="M42 (Orion Nebula)",
            ),
            StartImagingCommand(
                name="Start imaging",
                exposure_time=600.0,  # 10 minute exposures
                gain=120,
            ),
            WaitUntilEventCommand(
                name="Wait until dawn",
                event=AstronomicalEvent.ASTRONOMICAL_DAWN,
                latitude=latitude,
                longitude=longitude,
            ),
            StopImagingCommand(name="Stop imaging"),
        ],
    )


def create_multi_target_sequence() -> Sequence:
    """Create a sequence that images multiple targets.

    This sequence uses nested SequenceCommand to organize steps.

    Returns:
        Sequence: A configured multi-target sequence
    """
    # Helper function to create a target imaging sub-sequence
    def create_target_sequence(name: str, ra: float, dec: float, duration_minutes: float) -> SequenceCommand:
        return SequenceCommand(
            name=f"Image {name}",
            description=f"Complete imaging sequence for {name}",
            commands=[
                GoToTargetCommand(
                    name=f"Slew to {name}",
                    ra=ra,
                    dec=dec,
                    target_name=name,
                ),
                StartImagingCommand(
                    name=f"Start imaging {name}",
                    exposure_time=300.0,
                    gain=100,
                ),
                WaitMinutesCommand(
                    name=f"Image {name} for {duration_minutes} minutes",
                    minutes=duration_minutes,
                ),
                StopImagingCommand(name=f"Stop imaging {name}"),
            ],
        )

    return Sequence(
        name="Multi-Target Imaging",
        description="Image multiple deep sky objects in sequence",
        commands=[
            WaitUntilEventCommand(
                name="Wait for astronomical dusk",
                event=AstronomicalEvent.ASTRONOMICAL_DUSK,
                latitude=37.7749,
                longitude=-122.4194,
            ),
            create_target_sequence("M31 (Andromeda)", 10.6847, 41.2689, 90),
            create_target_sequence("M42 (Orion Nebula)", 83.8221, -5.3911, 60),
            create_target_sequence("M45 (Pleiades)", 56.75, 24.1167, 45),
        ],
    )


def create_timed_sequence() -> Sequence:
    """Create a sequence that starts at a specific time.

    Returns:
        Sequence: A configured timed sequence
    """
    # Start imaging at 10 PM tonight
    tonight_10pm = datetime.now().replace(hour=22, minute=0, second=0, microsecond=0)
    if tonight_10pm < datetime.now():
        tonight_10pm += timedelta(days=1)

    return Sequence(
        name="Timed Imaging Session",
        description="Start imaging at a specific time",
        commands=[
            WaitUntilTimeCommand(
                name="Wait until 10 PM",
                target_time=tonight_10pm,
            ),
            GoToTargetCommand(
                name="Slew to NGC 7000",
                ra=312.25,
                dec=44.5,
                target_name="NGC 7000 (North America Nebula)",
            ),
            StartImagingCommand(
                name="Start imaging",
                exposure_time=600.0,
                gain=100,
                count=30,  # 30 x 10min = 5 hours
            ),
            WaitMinutesCommand(name="Image for 5 hours", minutes=300),
            StopImagingCommand(name="Stop imaging"),
        ],
    )


# Example of saving and loading sequences
if __name__ == "__main__":
    # Create a sequence
    seq = create_simple_imaging_sequence()

    # Save to JSON
    json_str = seq.to_json(file_path="simple_imaging.json")
    print("Saved sequence to simple_imaging.json")
    print(f"\nJSON:\n{json_str}")

    # Load from JSON
    loaded_seq = Sequence.from_json(file_path="simple_imaging.json")
    print(f"\nLoaded sequence: {loaded_seq.name}")
    print(f"Number of commands: {len(loaded_seq.commands)}")



"""Translates Seestar events to unified V2 events.

This module bridges the existing Seestar event system to the unified
V2 event system, allowing V2 consumers to receive events in a
protocol-agnostic format.
"""

from typing import Any

from scopinator.seestar.events import (
    AutoGotoEvent,
    ContinuousExposureEvent,
    FocuserMoveEvent,
    PiStatusEvent,
    StackEvent,
    ScopeGotoEvent,
    BaseEvent,
)
from scopinator.util.eventbus import EventBus as SeestarEventBus
from scopinator.v2.core.events import (
    EventType,
    ExposureEvent,
    FocuserEvent,
    SlewEvent,
    UnifiedEvent,
    UnifiedEventBus,
)
from scopinator.v2.core.types import Coordinates, SlewState


class SeestarEventTranslator:
    """Bridges Seestar events to the unified V2 event system.

    This translator subscribes to events from the existing Seestar
    EventBus and translates them to UnifiedEvent instances, which
    are then emitted on the V2 UnifiedEventBus.

    Example:
        translator = SeestarEventTranslator(unified_bus)

        # Pass translator.seestar_event_bus to SeestarClient
        client = SeestarClient(
            host="192.168.1.100",
            event_bus=translator.seestar_event_bus
        )
    """

    def __init__(self, unified_event_bus: UnifiedEventBus) -> None:
        """Initialize translator.

        Args:
            unified_event_bus: The V2 event bus to emit translated events to
        """
        self._unified_bus = unified_event_bus

        # Create a Seestar event bus for the client to use
        self._seestar_bus = SeestarEventBus()

        # Subscribe to Seestar events
        self._seestar_bus.add_listener("AutoGoto", self._handle_autogoto)
        self._seestar_bus.add_listener("ScopeGoto", self._handle_scope_goto)
        self._seestar_bus.add_listener("Stack", self._handle_stack)
        self._seestar_bus.add_listener("ContinuousExposure", self._handle_continuous_exposure)
        self._seestar_bus.add_listener("FocuserMove", self._handle_focuser)
        self._seestar_bus.add_listener("PiStatus", self._handle_status)

    @property
    def seestar_event_bus(self) -> SeestarEventBus:
        """Get the Seestar event bus to pass to SeestarClient."""
        return self._seestar_bus

    async def _handle_autogoto(self, event: AutoGotoEvent) -> None:
        """Translate AutoGoto events to SlewEvents."""
        if event.state == "start":
            unified = SlewEvent(
                event_type=EventType.SLEW_STARTED,
                source_device="seestar_mount",
                source_backend="seestar",
                state=SlewState.SLEWING,
            )
        elif event.state == "complete":
            unified = SlewEvent(
                event_type=EventType.SLEW_COMPLETED,
                source_device="seestar_mount",
                source_backend="seestar",
                state=SlewState.TRACKING,
            )
        elif event.state in ("cancel", "fail"):
            unified = SlewEvent(
                event_type=EventType.SLEW_ABORTED,
                source_device="seestar_mount",
                source_backend="seestar",
                state=SlewState.ERROR if event.state == "fail" else SlewState.IDLE,
                data={"error": event.error, "code": event.code},
            )
        else:
            # "working" state - emit progress event
            unified = SlewEvent(
                event_type=EventType.SLEW_PROGRESS,
                source_device="seestar_mount",
                source_backend="seestar",
                state=SlewState.SLEWING,
                data={"lapse_ms": event.lapse_ms},
            )

        await self._unified_bus.emit(unified)

    async def _handle_scope_goto(self, event: ScopeGotoEvent) -> None:
        """Translate ScopeGoto progress events to SlewEvents."""
        coords = None
        if event.cur_ra_dec:
            # Convert RA from hours to degrees
            coords = Coordinates(
                ra=event.cur_ra_dec.ra * 15.0,
                dec=event.cur_ra_dec.dec,
            )

        unified = SlewEvent(
            event_type=EventType.COORDINATES_UPDATED,
            source_device="seestar_mount",
            source_backend="seestar",
            current_coordinates=coords,
            state=SlewState.SLEWING,
            distance_remaining=event.dist_deg,
            data={"percent": getattr(event, "percent", None)},
        )
        await self._unified_bus.emit(unified)

    async def _handle_stack(self, event: StackEvent) -> None:
        """Translate Stack events to ExposureEvents."""
        if event.state == "start":
            event_type = EventType.STACK_STARTED
        elif event.state == "frame_complete":
            event_type = EventType.STACK_FRAME
        elif event.state == "complete":
            event_type = EventType.STACK_COMPLETED
        elif event.state == "cancel":
            event_type = EventType.EXPOSURE_ABORTED
        else:
            event_type = EventType.EXPOSURE_PROGRESS

        unified = ExposureEvent(
            event_type=event_type,
            source_device="seestar_camera",
            source_backend="seestar",
            frames_stacked=event.stacked_frame,
            frames_dropped=event.dropped_frame,
            data={
                "exp_ms": event.exp_ms,
                "dead_time_ms": event.dead_time_ms,
                "reject_type": event.reject_type,
            },
        )
        await self._unified_bus.emit(unified)

    async def _handle_continuous_exposure(self, event: ContinuousExposureEvent) -> None:
        """Translate ContinuousExposure events to ExposureEvents."""
        if event.state == "start":
            event_type = EventType.EXPOSURE_STARTED
        elif event.state == "complete":
            event_type = EventType.EXPOSURE_COMPLETED
        elif event.state == "cancel":
            event_type = EventType.EXPOSURE_ABORTED
        else:
            event_type = EventType.EXPOSURE_PROGRESS

        unified = ExposureEvent(
            event_type=event_type,
            source_device="seestar_camera",
            source_backend="seestar",
            data={"fps": event.fps, "lapse_ms": event.lapse_ms},
        )
        await self._unified_bus.emit(unified)

    async def _handle_focuser(self, event: FocuserMoveEvent) -> None:
        """Translate FocuserMove events to FocuserEvents."""
        if event.state == "complete":
            event_type = EventType.FOCUSER_STOPPED
        elif event.state == "start":
            event_type = EventType.FOCUSER_MOVING
        else:
            event_type = EventType.FOCUSER_POSITION_CHANGED

        unified = FocuserEvent(
            event_type=event_type,
            source_device="seestar_focuser",
            source_backend="seestar",
            position=event.position,
        )
        await self._unified_bus.emit(unified)

    async def _handle_status(self, event: PiStatusEvent) -> None:
        """Translate PiStatus events to generic status updates."""
        unified = UnifiedEvent(
            event_type=EventType.STATUS_UPDATE,
            source_device="seestar",
            source_backend="seestar",
            data={
                "temperature": event.temp,
                "battery_capacity": event.battery_capacity,
                "charger_status": event.charger_status,
                "charge_online": event.charge_online,
            },
        )
        await self._unified_bus.emit(unified)



"""Seestar mount implementation.

Wraps the existing SeestarClient to provide the V2 Mount interface.
"""

from typing import TYPE_CHECKING, Optional

from scopinator.seestar.commands.simple import ScopeGetEquCoord, ScopePark, ScopeSync
from scopinator.v2.core.capabilities import MountCapabilities
from scopinator.v2.core.devices import Mount, MountStatus
from scopinator.v2.core.events import EventType, SlewEvent, UnifiedEventBus
from scopinator.v2.core.exceptions import DeviceError, NotConnectedError
from scopinator.v2.core.types import Coordinates, PierSide, SlewState, TrackingRate

if TYPE_CHECKING:
    from scopinator.seestar.client import SeestarClient


class SeestarMount(Mount):
    """Mount implementation for Seestar using existing client.

    This is an adapter that wraps the existing SeestarClient to provide
    the V2 Mount interface. It delegates all operations to the underlying
    client while translating data types and emitting unified events.
    """

    def __init__(
        self,
        client: "SeestarClient",
        event_bus: Optional[UnifiedEventBus] = None,
    ) -> None:
        """Initialize Seestar mount adapter.

        Args:
            client: Existing SeestarClient instance (should already be connected)
            event_bus: Optional event bus for emitting mount events
        """
        super().__init__(event_bus)
        self._client = client

    async def connect(self) -> None:
        """Mount connection is handled by backend/client."""
        self._connected = self._client.is_connected

    async def disconnect(self) -> None:
        """Mount disconnection is handled by backend/client."""
        pass

    async def get_capabilities(self) -> MountCapabilities:
        """Get Seestar mount capabilities."""
        return MountCapabilities(
            can_slew=True,
            can_slew_async=True,
            can_sync=True,
            can_park=True,
            can_unpark=False,  # Seestar doesn't support unpark
            can_find_home=False,
            can_pulse_guide=False,
            can_set_tracking=True,  # Implicitly via start/stop view
            can_set_tracking_rate=False,
            can_slew_altaz=False,
            has_pier_side=False,
            alignment_mode="altaz",
            max_slew_rate=4.0,  # Approximate
        )

    async def get_coordinates(self) -> Coordinates:
        """Get current coordinates from Seestar."""
        if not self._client.is_connected:
            raise NotConnectedError("Seestar is not connected")

        response = await self._client.send_and_recv(ScopeGetEquCoord())
        if response and response.result:
            # Seestar returns RA in hours (0-24), convert to degrees (0-360)
            ra_hours = response.result.get("ra", 0)
            dec = response.result.get("dec", 0)
            return Coordinates(ra=ra_hours * 15.0, dec=dec)

        # Fallback to status if command fails
        if self._client.status.ra is not None:
            return Coordinates(
                ra=self._client.status.ra * 15.0,
                dec=self._client.status.dec or 0.0,
            )

        raise DeviceError("Failed to get coordinates from Seestar")

    async def slew_to_coordinates(
        self,
        coords: Coordinates,
        *,
        wait: bool = True,
    ) -> None:
        """Slew to coordinates.

        Args:
            coords: Target coordinates (RA in degrees)
            wait: If True, wait for slew to complete
        """
        if not self._client.is_connected:
            raise NotConnectedError("Seestar is not connected")

        # Emit slew started event
        event = SlewEvent(
            event_type=EventType.SLEW_STARTED,
            source_device="seestar_mount",
            source_backend="seestar",
            target_coordinates=coords,
            state=SlewState.SLEWING,
        )
        self._event_bus.emit_nowait(event)

        # Use existing goto method - it expects RA in degrees
        # but the internal method handles conversion
        await self._client.goto(
            target_name="V2 Target",
            in_ra=coords.ra,  # Degrees
            in_dec=coords.dec,
            mode="star",
        )

        if wait:
            # Wait for AutoGoto event to complete
            success, error = await self._client.wait_for_event_completion(
                "AutoGoto", timeout=120.0
            )

            if success:
                event = SlewEvent(
                    event_type=EventType.SLEW_COMPLETED,
                    source_device="seestar_mount",
                    source_backend="seestar",
                    target_coordinates=coords,
                    state=SlewState.TRACKING,
                )
            else:
                event = SlewEvent(
                    event_type=EventType.SLEW_ABORTED,
                    source_device="seestar_mount",
                    source_backend="seestar",
                    target_coordinates=coords,
                    state=SlewState.ERROR,
                    data={"error": error},
                )

            self._event_bus.emit_nowait(event)

            if not success:
                raise DeviceError(f"Slew failed: {error}")

    async def abort_slew(self) -> None:
        """Abort current slew."""
        if not self._client.is_connected:
            raise NotConnectedError("Seestar is not connected")

        await self._client.stop_goto()

        event = SlewEvent(
            event_type=EventType.SLEW_ABORTED,
            source_device="seestar_mount",
            source_backend="seestar",
            state=SlewState.IDLE,
        )
        self._event_bus.emit_nowait(event)

    async def sync_to_coordinates(self, coords: Coordinates) -> None:
        """Sync mount to coordinates."""
        if not self._client.is_connected:
            raise NotConnectedError("Seestar is not connected")

        # Convert from degrees to hours for Seestar
        # ScopeSync expects params as (ra_hours, dec_degrees) tuple
        ra_hours = coords.ra / 15.0
        await self._client.send_and_recv(
            ScopeSync(params=(ra_hours, coords.dec))
        )

    async def park(self) -> None:
        """Park the Seestar."""
        if not self._client.is_connected:
            raise NotConnectedError("Seestar is not connected")

        await self._client.send_and_recv(ScopePark())

    async def set_tracking(self, enabled: bool) -> None:
        """Enable/disable tracking.

        Seestar manages tracking automatically - it tracks when observing
        and stops when idle. This method is a no-op.
        """
        # Seestar handles tracking implicitly
        pass

    async def get_tracking(self) -> bool:
        """Get current tracking state."""
        # Seestar tracks when in observation mode
        return self._client.client_mode in ("Stack", "ContinuousExposure", "AutoGoto")

    async def get_status(self) -> MountStatus:
        """Get comprehensive mount status."""
        try:
            coords = await self.get_coordinates()
        except Exception:
            coords = None

        tracking = await self.get_tracking()

        # Map client mode to slew state
        client_mode = self._client.client_mode
        if client_mode == "AutoGoto":
            state = SlewState.SLEWING
        elif client_mode in ("Stack", "ContinuousExposure"):
            state = SlewState.TRACKING
        elif client_mode == "Idle":
            state = SlewState.IDLE
        else:
            state = SlewState.IDLE

        return MountStatus(
            connected=self._client.is_connected,
            name="Seestar Mount",
            driver_info="Seestar V2 Adapter",
            driver_version="1.0.0",
            coordinates=coords,
            state=state,
            tracking=tracking,
            tracking_rate=TrackingRate.SIDEREAL,
            pier_side=PierSide.UNKNOWN,
            at_park=False,
            at_home=False,
            slewing=client_mode == "AutoGoto",
            target_coordinates=None,  # Could be extracted from status
        )



"""INDI backend - pyindi-client wrapper for INDI protocol support.

INDI (Instrument-Neutral Distributed Interface) is an open protocol
for controlling astronomy equipment, primarily used on Linux and macOS.

This backend requires the pyindi-client package:
    pip install pyindi-client

See: https://indilib.org/
"""

try:
    from scopinator.v2.backends.indi.backend import INDIBackend
    from scopinator.v2.backends.indi.mount import INDIMount
    from scopinator.v2.backends.indi.camera import INDICamera

    INDI_AVAILABLE = True
    __all__ = [
        "INDIBackend",
        "INDIMount",
        "INDICamera",
        "INDI_AVAILABLE",
    ]
except ImportError:
    INDI_AVAILABLE = False
    __all__ = ["INDI_AVAILABLE"]



"""ASCOM Alpaca focuser implementation.

Implements the V2 Focuser interface using the Alpaca HTTP REST API
for focuser control.

API Reference: https://ascom-standards.org/api/
"""

import asyncio
from typing import Callable, Optional

import aiohttp

from scopinator.v2.core.capabilities import FocuserCapabilities
from scopinator.v2.core.devices import Focuser, FocuserStatus
from scopinator.v2.core.events import EventType, UnifiedEvent, UnifiedEventBus
from scopinator.v2.core.exceptions import CommandError
from scopinator.v2.core.types import FocuserPosition


class AlpacaFocuser(Focuser):
    """Focuser implementation using ASCOM Alpaca REST API.

    All Alpaca focuser endpoints follow the pattern:
    GET/PUT /api/v1/focuser/{device_number}/{property}
    """

    def __init__(
        self,
        session: aiohttp.ClientSession,
        base_url: str,
        device_number: int,
        client_id: int,
        get_transaction_id: Callable[[], int],
        event_bus: Optional[UnifiedEventBus] = None,
    ) -> None:
        """Initialize Alpaca focuser.

        Args:
            session: aiohttp session for HTTP requests
            base_url: Base URL of Alpaca server
            device_number: Alpaca device number
            client_id: Alpaca client ID
            get_transaction_id: Function to get next transaction ID
            event_bus: Optional event bus for focuser events
        """
        super().__init__(event_bus)
        self._session = session
        self._base_url = f"{base_url}/api/v1/focuser/{device_number}"
        self._client_id = client_id
        self._get_tid = get_transaction_id
        self._timeout = aiohttp.ClientTimeout(total=30)

    async def _get(self, endpoint: str) -> dict:
        """Make GET request to Alpaca endpoint."""
        params = {
            "ClientID": self._client_id,
            "ClientTransactionID": self._get_tid(),
        }
        async with self._session.get(
            f"{self._base_url}/{endpoint}",
            params=params,
            timeout=self._timeout,
        ) as resp:
            data = await resp.json()
            self._check_error(data)
            return data

    async def _put(self, endpoint: str, form_data: Optional[dict] = None) -> dict:
        """Make PUT request to Alpaca endpoint."""
        data = form_data or {}
        data["ClientID"] = self._client_id
        data["ClientTransactionID"] = self._get_tid()

        async with self._session.put(
            f"{self._base_url}/{endpoint}",
            data=data,
            timeout=self._timeout,
        ) as resp:
            result = await resp.json()
            self._check_error(result)
            return result

    def _check_error(self, data: dict) -> None:
        """Check Alpaca response for errors."""
        error_number = data.get("ErrorNumber", 0)
        if error_number != 0:
            error_message = data.get("ErrorMessage", "Unknown error")
            raise CommandError(error_message, code=error_number)

    async def connect(self) -> None:
        """Connect to the focuser."""
        await self._put("connected", {"Connected": "true"})
        self._connected = True

    async def disconnect(self) -> None:
        """Disconnect from the focuser."""
        try:
            await self._put("connected", {"Connected": "false"})
        finally:
            self._connected = False

    async def get_capabilities(self) -> FocuserCapabilities:
        """Query focuser capabilities."""
        results = await asyncio.gather(
            self._get("maxstep"),
            self._get("stepsize"),
            self._get("tempcompavailable"),
            self._get("tempcomp"),
            return_exceptions=True,
        )

        def get_value(result, default):
            if isinstance(result, Exception):
                return default
            return result.get("Value", default)

        max_step = get_value(results[0], 0)
        step_size = get_value(results[1], 0.0)
        temp_comp_available = get_value(results[2], False)

        # Check if temperature is available by trying to read it
        has_temperature = False
        try:
            temp_result = await self._get("temperature")
            if temp_result.get("Value") is not None:
                has_temperature = True
        except Exception:
            pass

        return FocuserCapabilities(
            can_absolute=True,
            can_relative=True,
            can_halt=True,
            has_temperature=has_temperature,
            has_temp_compensation=temp_comp_available,
            max_step=max_step,
            step_size=float(step_size),
        )

    async def get_position(self) -> FocuserPosition:
        """Get current focuser position."""
        results = await asyncio.gather(
            self._get("position"),
            self._get("maxstep"),
            self._get("ismoving"),
            return_exceptions=True,
        )

        def get_value(result, default):
            if isinstance(result, Exception):
                return default
            return result.get("Value", default)

        position = get_value(results[0], 0)
        max_position = get_value(results[1], 1)
        is_moving = get_value(results[2], False)

        # Try to get temperature
        temperature = None
        try:
            temp_result = await self._get("temperature")
            temperature = temp_result.get("Value")
        except Exception:
            pass

        return FocuserPosition(
            position=position,
            max_position=max_position,
            temperature=temperature,
            is_moving=is_moving,
        )

    async def move_to(self, position: int, *, wait: bool = True) -> None:
        """Move to absolute position.

        Args:
            position: Target position in steps
            wait: If True, wait for move to complete
        """
        # Validate position
        max_step = (await self._get("maxstep"))["Value"]
        if position < 0 or position > max_step:
            raise ValueError(f"Position {position} out of range (0-{max_step})")

        await self._put("move", {"Position": str(position)})

        if wait:
            # Poll until movement is complete
            while True:
                result = await self._get("ismoving")
                if not result["Value"]:
                    break
                await asyncio.sleep(0.2)

            # Emit event
            event = UnifiedEvent(
                event_type=EventType.STATUS_UPDATE,
                source_device="alpaca_focuser",
                source_backend="alpaca",
                data={"position": position, "action": "move_complete"},
            )
            self._event_bus.emit_nowait(event)

    async def move_relative(self, steps: int, *, wait: bool = True) -> None:
        """Move relative to current position.

        Args:
            steps: Number of steps (positive = out, negative = in)
            wait: If True, wait for move to complete
        """
        current_pos = (await self._get("position"))["Value"]
        new_pos = current_pos + steps
        await self.move_to(new_pos, wait=wait)

    async def halt(self) -> None:
        """Halt any movement."""
        await self._put("halt")

    async def get_status(self) -> FocuserStatus:
        """Get current focuser status."""
        results = await asyncio.gather(
            self._get("position"),
            self._get("maxstep"),
            self._get("ismoving"),
            return_exceptions=True,
        )

        def get_value(result, default):
            if isinstance(result, Exception):
                return default
            return result.get("Value", default)

        position = get_value(results[0], 0)
        max_position = get_value(results[1], 0)
        is_moving = get_value(results[2], False)

        # Try to get temperature and temp comp status
        temperature = None
        temp_comp_enabled = False
        try:
            temp_result = await self._get("temperature")
            temperature = temp_result.get("Value")
        except Exception:
            pass
        try:
            temp_comp_result = await self._get("tempcomp")
            temp_comp_enabled = temp_comp_result.get("Value", False)
        except Exception:
            pass

        return FocuserStatus(
            connected=self._connected,
            name="Alpaca Focuser",
            driver_info="ASCOM Alpaca",
            driver_version="1.0.0",
            position=position,
            max_position=max_position,
            is_moving=is_moving,
            temperature=temperature,
            temp_comp_enabled=temp_comp_enabled,
        )



"""Unified event system for V2.

This module provides a protocol-agnostic event system that normalizes
events from different backends (Seestar, ASCOM Alpaca, INDI) into a
common format.
"""

import asyncio
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Coroutine, Optional

from pydantic import BaseModel, Field

from scopinator.v2.core.types import Coordinates, SlewState


class EventType(str, Enum):
    """Unified event types."""

    # Connection events
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    CONNECTION_ERROR = "connection_error"
    RECONNECTING = "reconnecting"

    # Mount events
    SLEW_STARTED = "slew_started"
    SLEW_PROGRESS = "slew_progress"
    SLEW_COMPLETED = "slew_completed"
    SLEW_ABORTED = "slew_aborted"
    TRACKING_CHANGED = "tracking_changed"
    COORDINATES_UPDATED = "coordinates_updated"
    PARK_STARTED = "park_started"
    PARK_COMPLETED = "park_completed"
    UNPARK_COMPLETED = "unpark_completed"

    # Camera events
    EXPOSURE_STARTED = "exposure_started"
    EXPOSURE_PROGRESS = "exposure_progress"
    EXPOSURE_COMPLETED = "exposure_completed"
    EXPOSURE_ABORTED = "exposure_aborted"
    IMAGE_READY = "image_ready"
    DOWNLOAD_PROGRESS = "download_progress"
    COOLER_CHANGED = "cooler_changed"

    # Stacking events (for smart scopes)
    STACK_STARTED = "stack_started"
    STACK_FRAME = "stack_frame"
    STACK_COMPLETED = "stack_completed"

    # Focuser events
    FOCUSER_MOVING = "focuser_moving"
    FOCUSER_STOPPED = "focuser_stopped"
    FOCUSER_POSITION_CHANGED = "focuser_position_changed"

    # Filter wheel events
    FILTER_CHANGING = "filter_changing"
    FILTER_CHANGED = "filter_changed"

    # General events
    STATUS_UPDATE = "status_update"
    ERROR = "error"
    WARNING = "warning"
    INFO = "info"


class UnifiedEvent(BaseModel):
    """Base unified event."""

    event_type: EventType
    timestamp: datetime = Field(default_factory=datetime.utcnow)
    source_device: str = Field(default="", description="Device that generated the event")
    source_backend: str = Field(default="", description="Backend type (seestar, alpaca, indi)")
    data: dict[str, Any] = Field(default_factory=dict, description="Event-specific data")

    model_config = {"use_enum_values": False}


class SlewEvent(UnifiedEvent):
    """Slew-related event with coordinate information."""

    target_coordinates: Optional[Coordinates] = None
    current_coordinates: Optional[Coordinates] = None
    state: SlewState = SlewState.IDLE
    distance_remaining: Optional[float] = Field(
        None, description="Remaining distance in degrees"
    )


class ExposureEvent(UnifiedEvent):
    """Exposure-related event with progress information."""

    duration_seconds: float = Field(default=0.0, description="Total exposure duration")
    elapsed_seconds: float = Field(default=0.0, description="Elapsed time")
    progress: float = Field(default=0.0, ge=0.0, le=1.0, description="Progress 0.0 to 1.0")
    frames_captured: int = Field(default=0, description="Number of frames captured")
    frames_stacked: int = Field(default=0, description="Number of frames stacked")
    frames_dropped: int = Field(default=0, description="Number of frames dropped")


class FocuserEvent(UnifiedEvent):
    """Focuser-related event with position information."""

    position: int = Field(default=0, description="Current position")
    target_position: Optional[int] = Field(None, description="Target position if moving")
    temperature: Optional[float] = Field(None, description="Temperature in Celsius")


class FilterEvent(UnifiedEvent):
    """Filter wheel-related event."""

    position: int = Field(default=0, description="Filter position (0-indexed)")
    filter_name: Optional[str] = Field(None, description="Filter name")
    previous_position: Optional[int] = Field(None, description="Previous position")


# Type alias for event handlers
EventHandler = Callable[[UnifiedEvent], Coroutine[Any, Any, None]]


class UnifiedEventBus:
    """Unified event bus that normalizes events from different backends.

    This event bus provides a common interface for subscribing to and
    emitting events, regardless of the underlying protocol.

    Example:
        bus = UnifiedEventBus()

        async def on_slew(event: SlewEvent):
            print(f"Slewing to {event.target_coordinates}")

        bus.subscribe(EventType.SLEW_STARTED, on_slew)
        await bus.emit(SlewEvent(event_type=EventType.SLEW_STARTED, ...))
    """

    def __init__(self) -> None:
        self._handlers: dict[EventType, set[EventHandler]] = {}
        self._global_handlers: set[EventHandler] = set()
        self._lock = asyncio.Lock()

    def subscribe(self, event_type: EventType, handler: EventHandler) -> None:
        """Subscribe to a specific event type.

        Args:
            event_type: The type of event to subscribe to
            handler: Async function to call when event is emitted
        """
        if event_type not in self._handlers:
            self._handlers[event_type] = set()
        self._handlers[event_type].add(handler)

    def subscribe_all(self, handler: EventHandler) -> None:
        """Subscribe to all events.

        Args:
            handler: Async function to call for all events
        """
        self._global_handlers.add(handler)

    def unsubscribe(self, event_type: EventType, handler: EventHandler) -> None:
        """Unsubscribe from a specific event type.

        Args:
            event_type: The type of event to unsubscribe from
            handler: The handler to remove
        """
        if event_type in self._handlers:
            self._handlers[event_type].discard(handler)

    def unsubscribe_all(self, handler: EventHandler) -> None:
        """Unsubscribe from all events.

        Args:
            handler: The handler to remove
        """
        self._global_handlers.discard(handler)

    def clear(self) -> None:
        """Remove all handlers."""
        self._handlers.clear()
        self._global_handlers.clear()

    async def emit(self, event: UnifiedEvent) -> None:
        """Emit an event to all subscribers.

        Handlers are called concurrently. Exceptions in handlers are
        caught and logged but do not prevent other handlers from running.

        Args:
            event: The event to emit
        """
        handlers: set[EventHandler] = set(self._global_handlers)
        if event.event_type in self._handlers:
            handlers.update(self._handlers[event.event_type])

        if not handlers:
            return

        # Run all handlers concurrently, catching exceptions
        results = await asyncio.gather(
            *[self._safe_call(handler, event) for handler in handlers],
            return_exceptions=True,
        )

        # Log any exceptions (but don't raise)
        for result in results:
            if isinstance(result, Exception):
                # In production, this would use proper logging
                pass

    async def _safe_call(self, handler: EventHandler, event: UnifiedEvent) -> None:
        """Safely call a handler, catching exceptions."""
        try:
            await handler(event)
        except Exception:
            # Re-raise to be caught by gather
            raise

    def emit_nowait(self, event: UnifiedEvent) -> None:
        """Emit an event without waiting for handlers to complete.

        Creates an asyncio task to emit the event in the background.

        Args:
            event: The event to emit
        """
        try:
            loop = asyncio.get_running_loop()
            loop.create_task(self.emit(event))
        except RuntimeError:
            # No running loop - skip emission
            pass

    def handler_count(self, event_type: Optional[EventType] = None) -> int:
        """Get number of registered handlers.

        Args:
            event_type: If provided, count handlers for this type only.
                       If None, count all handlers.

        Returns:
            Number of registered handlers
        """
        if event_type is None:
            total = len(self._global_handlers)
            for handlers in self._handlers.values():
                total += len(handlers)
            return total
        return len(self._handlers.get(event_type, set())) + len(self._global_handlers)



"""INDI camera implementation.

Implements the V2 Camera interface using INDI CCD properties.
"""

import asyncio
from typing import TYPE_CHECKING, Optional

from scopinator.v2.core.capabilities import CameraCapabilities
from scopinator.v2.core.devices import Camera, CameraStatus
from scopinator.v2.core.events import EventType, ExposureEvent, UnifiedEventBus
from scopinator.v2.core.exceptions import DeviceError, NotConnectedError
from scopinator.v2.core.types import CameraState, ExposureSettings, ImageData

if TYPE_CHECKING:
    from scopinator.v2.backends.indi.backend import INDIClient

try:
    import PyIndi

    INDI_AVAILABLE = True
except ImportError:
    INDI_AVAILABLE = False
    PyIndi = None


class INDICamera(Camera):
    """Camera implementation using INDI protocol.

    Key INDI properties for CCDs:
    - CCD_EXPOSURE: Exposure control (number)
    - CCD_FRAME: Frame size (number)
    - CCD_BINNING: Binning (number)
    - CCD1: Image data (blob)
    - CCD_TEMPERATURE: Cooler temperature (number)
    - CCD_COOLER: Cooler on/off (switch)
    """

    def __init__(
        self,
        client: "INDIClient",
        device_name: str,
        event_bus: Optional[UnifiedEventBus] = None,
    ) -> None:
        """Initialize INDI camera.

        Args:
            client: INDI client connection
            device_name: INDI device name
            event_bus: Optional event bus for camera events
        """
        super().__init__(event_bus)
        self._client = client
        self._device_name = device_name
        self._device: Optional["PyIndi.BaseDevice"] = None
        self._last_blob: Optional[bytes] = None
        self._last_blob_width: int = 0
        self._last_blob_height: int = 0

    def _get_device(self) -> "PyIndi.BaseDevice":
        """Get the INDI device."""
        if self._device is None:
            self._device = self._client.get_device(self._device_name)
        if self._device is None:
            raise NotConnectedError(f"Device {self._device_name} not found")
        return self._device

    def _get_number(self, property_name: str) -> Optional["PyIndi.INumberVectorProperty"]:
        """Get a number property."""
        device = self._get_device()
        return device.getNumber(property_name)

    def _get_switch(self, property_name: str) -> Optional["PyIndi.ISwitchVectorProperty"]:
        """Get a switch property."""
        device = self._get_device()
        return device.getSwitch(property_name)

    async def connect(self) -> None:
        """Connect to the camera."""
        device = self._get_device()

        connection = device.getSwitch("CONNECTION")
        if connection:
            for i in range(connection.nsp):
                if connection[i].name == "CONNECT":
                    connection[i].s = PyIndi.ISS_ON
                else:
                    connection[i].s = PyIndi.ISS_OFF
            self._client.sendNewSwitch(connection)
            await asyncio.sleep(1.0)

        # Enable blob reception
        self._client.setBLOBMode(PyIndi.B_ALSO, self._device_name, None)

        self._connected = True

    async def disconnect(self) -> None:
        """Disconnect from the camera."""
        try:
            device = self._get_device()
            connection = device.getSwitch("CONNECTION")
            if connection:
                for i in range(connection.nsp):
                    if connection[i].name == "DISCONNECT":
                        connection[i].s = PyIndi.ISS_ON
                    else:
                        connection[i].s = PyIndi.ISS_OFF
                self._client.sendNewSwitch(connection)
        finally:
            self._connected = False

    async def get_capabilities(self) -> CameraCapabilities:
        """Get camera capabilities."""
        device = self._get_device()

        # Get frame info
        frame = self._get_number("CCD_FRAME")
        width = 0
        height = 0
        if frame:
            for i in range(frame.nnp):
                if frame[i].name == "WIDTH":
                    width = int(frame[i].value)
                elif frame[i].name == "HEIGHT":
                    height = int(frame[i].value)

        return CameraCapabilities(
            can_abort_exposure=True,
            can_set_gain=device.getNumber("CCD_GAIN") is not None,
            can_set_binning=device.getNumber("CCD_BINNING") is not None,
            can_subframe=frame is not None,
            has_cooler=device.getSwitch("CCD_COOLER") is not None,
            sensor_width=width,
            sensor_height=height,
        )

    async def start_exposure(self, settings: ExposureSettings) -> None:
        """Start an exposure."""
        # Set binning if available
        if settings.bin_x != 1 or settings.bin_y != 1:
            binning = self._get_number("CCD_BINNING")
            if binning:
                for i in range(binning.nnp):
                    if binning[i].name == "HOR_BIN":
                        binning[i].value = settings.bin_x
                    elif binning[i].name == "VER_BIN":
                        binning[i].value = settings.bin_y
                self._client.sendNewNumber(binning)

        # Emit exposure started event
        event = ExposureEvent(
            event_type=EventType.EXPOSURE_STARTED,
            source_device=self._device_name,
            source_backend="indi",
            duration_seconds=settings.duration_seconds,
        )
        self._event_bus.emit_nowait(event)

        # Start exposure
        exposure = self._get_number("CCD_EXPOSURE")
        if exposure is None:
            raise DeviceError("CCD_EXPOSURE property not available")

        for i in range(exposure.nnp):
            if exposure[i].name == "CCD_EXPOSURE_VALUE":
                exposure[i].value = settings.duration_seconds
        self._client.sendNewNumber(exposure)

    async def abort_exposure(self) -> None:
        """Abort current exposure."""
        abort = self._get_switch("CCD_ABORT_EXPOSURE")
        if abort:
            for i in range(abort.nsp):
                if abort[i].name == "ABORT":
                    abort[i].s = PyIndi.ISS_ON
            self._client.sendNewSwitch(abort)

        event = ExposureEvent(
            event_type=EventType.EXPOSURE_ABORTED,
            source_device=self._device_name,
            source_backend="indi",
        )
        self._event_bus.emit_nowait(event)

    async def is_exposing(self) -> bool:
        """Check if currently exposing."""
        exposure = self._get_number("CCD_EXPOSURE")
        if exposure:
            return exposure.s == PyIndi.IPS_BUSY
        return False

    async def get_image(self) -> ImageData:
        """Get the last captured image.

        Note: INDI sends images as BLOBs which need to be caught
        by a callback. This is a simplified implementation.
        """
        if self._last_blob is None:
            raise DeviceError("No image available")

        return ImageData(
            width=self._last_blob_width or 1,
            height=self._last_blob_height or 1,
            data=self._last_blob,
            bit_depth=16,
            is_color=False,
        )

    async def set_cooler(
        self, enabled: bool, setpoint: Optional[float] = None
    ) -> None:
        """Control cooler."""
        cooler = self._get_switch("CCD_COOLER")
        if cooler:
            for i in range(cooler.nsp):
                if enabled and cooler[i].name == "COOLER_ON":
                    cooler[i].s = PyIndi.ISS_ON
                elif not enabled and cooler[i].name == "COOLER_OFF":
                    cooler[i].s = PyIndi.ISS_ON
                else:
                    cooler[i].s = PyIndi.ISS_OFF
            self._client.sendNewSwitch(cooler)

        if setpoint is not None and enabled:
            temp = self._get_number("CCD_TEMPERATURE")
            if temp:
                for i in range(temp.nnp):
                    if temp[i].name == "CCD_TEMPERATURE_VALUE":
                        temp[i].value = setpoint
                self._client.sendNewNumber(temp)

    async def get_temperature(self) -> Optional[float]:
        """Get sensor temperature."""
        temp = self._get_number("CCD_TEMPERATURE")
        if temp:
            for i in range(temp.nnp):
                if temp[i].name == "CCD_TEMPERATURE_VALUE":
                    return temp[i].value
        return None

    async def get_status(self) -> CameraStatus:
        """Get camera status."""
        exposure = self._get_number("CCD_EXPOSURE")
        exposing = exposure is not None and exposure.s == PyIndi.IPS_BUSY

        temp = await self.get_temperature()

        cooler = self._get_switch("CCD_COOLER")
        cooler_on = False
        if cooler:
            for i in range(cooler.nsp):
                if cooler[i].name == "COOLER_ON" and cooler[i].s == PyIndi.ISS_ON:
                    cooler_on = True
                    break

        state = CameraState.EXPOSING if exposing else CameraState.IDLE

        return CameraStatus(
            connected=self._connected,
            name=self._device_name,
            driver_info="INDI",
            driver_version="1.0.0",
            state=state,
            temperature=temp,
            cooler_on=cooler_on,
        )



"""Centralized logging configuration for scopinator using loguru.

Supports configuration via:
1. Environment variables: SCOPINATOR_LOG_LEVEL, SCOPINATOR_DEBUG, SCOPINATOR_TRACE
2. CLI flags: --debug, --trace, --quiet
3. Programmatic configuration

Log levels (loguru):
- TRACE (5): Most verbose, includes all internal state changes
- DEBUG (10): Detailed debugging information  
- INFO (20): General informational messages
- SUCCESS (25): Success messages (loguru specific)
- WARNING (30): Warning messages
- ERROR (40): Error messages
- CRITICAL (50): Critical errors
"""

import os
import sys
from typing import Optional
from loguru import logger

# Remove default logger
logger.remove()


class LoggingConfig:
    """Manage logging configuration for scopinator using loguru."""
    
    _initialized = False
    _current_mode = "default"
    _current_level = "INFO"
    
    @classmethod
    def configure(
        cls,
        debug: bool = False,
        trace: bool = False,
        quiet: bool = False,
        level: Optional[str] = None,
        force: bool = False,
    ) -> None:
        """Configure logging based on flags and environment variables.
        
        Args:
            debug: Enable debug logging
            trace: Enable trace logging (most verbose)
            quiet: Reduce logging to warnings and errors only
            level: Explicit log level (overrides other flags)
            force: Force reconfiguration even if already initialized
        """
        if cls._initialized and not force:
            return
        
        # Remove any existing handlers
        logger.remove()
        
        # Check environment variables (lowest priority)
        env_level = os.environ.get("SCOPINATOR_LOG_LEVEL", "").upper()
        env_debug = os.environ.get("SCOPINATOR_DEBUG", "").lower() in ("true", "1", "yes")
        env_trace = os.environ.get("SCOPINATOR_TRACE", "").lower() in ("true", "1", "yes")
        env_quiet = os.environ.get("SCOPINATOR_QUIET", "").lower() in ("true", "1", "yes")
        
        # Apply environment settings if not overridden by arguments
        if not any([debug, trace, quiet, level]):
            debug = env_debug
            trace = env_trace
            quiet = env_quiet
            if env_level:
                level = env_level
        
        # Determine log level
        if level:
            # Explicit level overrides everything
            log_level = level.upper()
            cls._current_mode = f"custom-{level.lower()}"
        elif trace:
            log_level = "TRACE"
            cls._current_mode = "trace"
        elif debug:
            log_level = "DEBUG"
            cls._current_mode = "debug"
        elif quiet:
            log_level = "WARNING"
            cls._current_mode = "quiet"
        else:
            log_level = "INFO"
            cls._current_mode = "default"
        
        cls._current_level = log_level
        
        # Configure format based on verbosity
        if trace or cls._current_mode == "trace":
            # Most detailed format for trace mode
            log_format = (
                "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
                "<level>{level: <8}</level> | "
                "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
                "<level>{message}</level>"
            )
        elif debug or cls._current_mode == "debug":
            # Detailed format for debug mode
            log_format = (
                "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | "
                "<level>{level: <8}</level> | "
                "<cyan>{name}</cyan> - "
                "<level>{message}</level>"
            )
        elif quiet or cls._current_mode == "quiet":
            # Minimal format for quiet mode
            log_format = "<level>{level}: {message}</level>"
        else:
            # Standard format for normal mode
            log_format = (
                "<green>{time:HH:mm:ss}</green> | "
                "<level>{level}</level> | "
                "<level>{message}</level>"
            )
        
        # Add console handler with configured format and level
        logger.add(
            sys.stderr,
            format=log_format,
            level=log_level,
            colorize=True,
            backtrace=(trace or cls._current_mode == "trace"),
            diagnose=(trace or cls._current_mode == "trace"),
        )
        
        # Add filter for specific modules based on mode
        if cls._current_mode == "quiet":
            # In quiet mode, suppress info from noisy modules
            logger.disable("scopinator.seestar.rtspclient")
            logger.disable("scopinator.util.eventbus")
        elif cls._current_mode not in ("debug", "trace"):
            # In normal mode, reduce verbosity of some modules
            def module_filter(record):
                # Reduce RTSP client noise
                if record["name"].startswith("scopinator.seestar.rtspclient"):
                    return record["level"].no >= logger.level("WARNING").no
                # Reduce eventbus noise
                if record["name"].startswith("scopinator.util.eventbus"):
                    return record["level"].no >= logger.level("WARNING").no
                return True
            
            logger.add(
                lambda msg: None,  # Null sink
                filter=module_filter,
                level=0
            )
        
        cls._initialized = True
        
        # Log the configuration (only if not quiet)
        if not quiet and cls._current_mode != "quiet":
            logger.debug(f"Logging configured: mode={cls._current_mode}, level={log_level}")
    
    @classmethod
    def get_current_mode(cls) -> str:
        """Get the current logging mode."""
        return cls._current_mode
    
    @classmethod
    def get_current_level(cls) -> str:
        """Get the current logging level."""
        return cls._current_level
    
    @classmethod
    def is_debug_enabled(cls) -> bool:
        """Check if debug logging is enabled."""
        return cls._current_mode in ("debug", "trace") or cls._current_level in ("DEBUG", "TRACE")
    
    @classmethod
    def is_trace_enabled(cls) -> bool:
        """Check if trace logging is enabled."""
        return cls._current_mode == "trace" or cls._current_level == "TRACE"
    
    @classmethod
    def reset(cls) -> None:
        """Reset logging configuration to allow reconfiguration."""
        logger.remove()
        cls._initialized = False
        cls._current_mode = "default"
        cls._current_level = "INFO"


def setup_logging(
    debug: bool = False,
    trace: bool = False,
    quiet: bool = False,
    level: Optional[str] = None,
) -> None:
    """Convenience function to set up logging.
    
    Args:
        debug: Enable debug logging
        trace: Enable trace logging (most verbose)
        quiet: Reduce logging to warnings and errors only
        level: Explicit log level string
    """
    LoggingConfig.configure(debug=debug, trace=trace, quiet=quiet, level=level)


def get_logger(name: Optional[str] = None):
    """Get the loguru logger instance.
    
    Args:
        name: Logger name (not used by loguru, kept for compatibility)
        
    Returns:
        The loguru logger instance
    """
    # Ensure logging is configured
    if not LoggingConfig._initialized:
        LoggingConfig.configure()
    
    # Loguru uses a single logger instance
    return logger


"""Astronomical events for wait conditions."""

from datetime import datetime
from enum import Enum
from typing import Optional

from astropy import units as u
from astropy.coordinates import EarthLocation, get_sun
from astropy.time import Time


class AstronomicalEvent(str, Enum):
    """Astronomical events that can be used as wait conditions."""

    ASTRONOMICAL_DAWN = "astronomical_dawn"
    NAUTICAL_DAWN = "nautical_dawn"
    CIVIL_DAWN = "civil_dawn"
    SUNRISE = "sunrise"
    SUNSET = "sunset"
    CIVIL_DUSK = "civil_dusk"
    NAUTICAL_DUSK = "nautical_dusk"
    ASTRONOMICAL_DUSK = "astronomical_dusk"


def calculate_event_time(
    event: AstronomicalEvent,
    location: EarthLocation,
    date: Optional[datetime] = None,
) -> datetime:
    """Calculate the time of an astronomical event.

    Args:
        event: The astronomical event
        location: Observer's location on Earth
        date: Date for which to calculate (defaults to today)

    Returns:
        datetime: Time when the event occurs

    Raises:
        ValueError: If the event doesn't occur on the given date
    """
    if date is None:
        date = datetime.utcnow()

    time = Time(date)

    # Define altitude thresholds for different events
    event_altitudes = {
        AstronomicalEvent.ASTRONOMICAL_DAWN: -18 * u.deg,
        AstronomicalEvent.NAUTICAL_DAWN: -12 * u.deg,
        AstronomicalEvent.CIVIL_DAWN: -6 * u.deg,
        AstronomicalEvent.SUNRISE: -0.833 * u.deg,  # Account for refraction and sun's radius
        AstronomicalEvent.SUNSET: -0.833 * u.deg,
        AstronomicalEvent.CIVIL_DUSK: -6 * u.deg,
        AstronomicalEvent.NAUTICAL_DUSK: -12 * u.deg,
        AstronomicalEvent.ASTRONOMICAL_DUSK: -18 * u.deg,
    }

    target_altitude = event_altitudes[event]

    # For dawn events, we need the time when the sun is rising
    # For dusk events, we need the time when the sun is setting
    is_dawn = event in [
        AstronomicalEvent.ASTRONOMICAL_DAWN,
        AstronomicalEvent.NAUTICAL_DAWN,
        AstronomicalEvent.CIVIL_DAWN,
        AstronomicalEvent.SUNRISE,
    ]

    # Search for the event time (simplified - in production use astropy's built-in functions)
    # This is a basic implementation - for production use astroplan library
    from astropy.coordinates import AltAz

    # Sample times throughout the day
    times = Time(date) + (range(0, 24 * 60, 5)) * u.minute
    sun_positions = get_sun(times)
    altaz_frame = AltAz(obstime=times, location=location)
    altitudes = sun_positions.transform_to(altaz_frame).alt

    # Find when altitude crosses the threshold
    for i in range(len(altitudes) - 1):
        current_alt = altitudes[i]
        next_alt = altitudes[i + 1]

        if is_dawn:
            # Dawn: altitude increasing through threshold
            if current_alt < target_altitude <= next_alt:
                return times[i].datetime
        else:
            # Dusk: altitude decreasing through threshold
            if current_alt > target_altitude >= next_alt:
                return times[i].datetime

    raise ValueError(f"Event {event} does not occur on {date.date()}")



"""CLI commands module."""

from scopinator.cli.commands.discovery import discover_telescopes
from scopinator.cli.commands.profile import (
    profile,
    list_profiles,
    get_profile_path,
    load_profile,
    get_current_profile,
)

__all__ = [
    "discover_telescopes",
    "profile",
    "list_profiles",
    "get_profile_path",
    "load_profile",
    "get_current_profile",
]


"""Establish connection with Seestar."""

import asyncio
from asyncio import StreamReader, StreamWriter, IncompleteReadError
from collections.abc import Callable
from typing import Optional
from pydantic import BaseModel, ConfigDict
from scopinator.util.logging_config import get_logger
logging = get_logger(__name__)
import random


class SeestarConnection(BaseModel, arbitrary_types_allowed=True):
    """Connection with Seestar."""
    model_config = ConfigDict(extra="allow")

    reader: StreamReader | None = None
    writer: StreamWriter | None = None
    host: str
    port: int
    written_messages: int = 0
    read_messages: int = 0
    _is_connected: bool = False
    _reconnect_attempts: int = 0
    _max_reconnect_attempts: int = 5
    _base_reconnect_delay: float = 1.0
    _max_reconnect_delay: float = 60.0
    connection_timeout: float = 10.0
    read_timeout: float = 30.0
    write_timeout: float = 10.0
    _should_reconnect_callback: Optional[Callable[[], bool]] = None
    _last_reconnect_log: float = 0.0
    _reconnect_log_interval: float = 30.0
    _reboot_detected: bool = False
    _last_reboot_time: float = 0.0
    _reconnect_lock: asyncio.Lock | None = None
    _reconnect_in_progress: bool = False

    def __str__(self) -> str:
        return f"{self.host}:{self.port}"

    def __repr__(self) -> str:
        return f"SeestarConnection(host={self.host!r}, port={self.port!r})"

    def __init__(
        self,
        host: str,
        port: int,
        connection_timeout: float = 10.0,
        read_timeout: float = 30.0,
        write_timeout: float = 10.0,
        should_reconnect_callback: Optional[Callable[[], bool]] = None,
        **kwargs,
    ):
        super().__init__(
            host=host,
            port=port,
            connection_timeout=connection_timeout,
            read_timeout=read_timeout,
            write_timeout=write_timeout,
            _should_reconnect_callback=should_reconnect_callback,
            **kwargs,
        )
        # Initialize the lock after super().__init__
        self._reconnect_lock = asyncio.Lock()

    async def open(self):
        """Open connection with Seestar."""
        try:
            self.reader, self.writer = await asyncio.wait_for(
                asyncio.open_connection(self.host, self.port),
                timeout=self.connection_timeout,
            )
            self._is_connected = True
            self._reconnect_attempts = 0
            self._last_reconnect_log = 0.0
            logging.info(
                f"Successfully connected to Seestar at {self.host}:{self.port}"
            )
        except asyncio.TimeoutError:
            self._is_connected = False
            logging.error(
                f"Connection timeout after {self.connection_timeout}s to {self.host}:{self.port}"
            )
            raise
        except Exception as e:
            self._is_connected = False
            logging.error(f"Failed to connect to Seestar: {e}")
            raise
        self.written_messages = 0
        self.read_messages = 0

    async def close(self):
        """Close connection with Seestar."""
        self._is_connected = False
        if self.writer:
            self.writer.close()
            try:
                await self.writer.wait_closed()
            except ConnectionResetError:
                # Connection already reset; ignore during cleanup
                pass
        self.reader = None
        self.writer = None

    def is_connected(self) -> bool:
        """Check if connection is established."""
        return (
            self._is_connected and self.reader is not None and self.writer is not None
        )
    
    def get_connection_stats(self) -> dict:
        """Get connection statistics."""
        return {
            "is_connected": self.is_connected(),
            "reconnect_attempts": self._reconnect_attempts,
            "max_reconnect_attempts": self._max_reconnect_attempts,
            "written_messages": self.written_messages,
            "read_messages": self.read_messages,
            "host": self.host,
            "port": self.port
        }

    def _is_connection_reset_error(self, error: Exception) -> bool:
        """Check if the error indicates a connection reset."""
        return isinstance(
            error,
            (
                ConnectionResetError,
                ConnectionAbortedError,
                BrokenPipeError,
                IncompleteReadError,
                OSError,
                asyncio.TimeoutError,
            ),
        )

    async def _reconnect_with_backoff(self) -> bool:
        """Attempt to reconnect with exponential backoff."""
        import time
        
        # Use lock to prevent concurrent reconnection attempts
        if self._reconnect_lock is None:
            self._reconnect_lock = asyncio.Lock()
            
        async with self._reconnect_lock:
            # Check if another coroutine already reconnected
            if self.is_connected():
                return True
            
            # Check if reconnection is already in progress
            if self._reconnect_in_progress:
                # Wait for the other reconnection attempt to complete
                max_wait = 60  # seconds
                start = time.time()
                while self._reconnect_in_progress and (time.time() - start) < max_wait:
                    await asyncio.sleep(0.5)
                    if self.is_connected():
                        return True
                return self.is_connected()
            
            self._reconnect_in_progress = True
            try:
                return await self._do_reconnect_with_backoff()
            finally:
                self._reconnect_in_progress = False
    
    async def _do_reconnect_with_backoff(self) -> bool:
        """Internal method to perform reconnection with backoff."""
        import time
        current_time = time.time()
        
        # Check if reconnection is allowed via callback
        if self._should_reconnect_callback and not self._should_reconnect_callback():
            logging.debug(
                f"Reconnection skipped for {self.host}:{self.port} due to callback check"
            )
            return False

        self._reconnect_attempts += 1
        
        # Check if this looks like a reboot (multiple failures within a short time)
        if not self._reboot_detected and self._reconnect_attempts >= 3:
            if current_time - self._last_reboot_time > 300:  # 5 minutes since last reboot
                self._reboot_detected = True
                self._last_reboot_time = current_time
                logging.info(
                    f"Telescope at {self.host}:{self.port} appears to be rebooting. "
                    f"Will continue reconnection attempts with reduced logging."
                )
        
        # Use exponential backoff but cap at max delay, don't give up after max attempts
        # Reset attempts counter periodically to prevent overflow and allow fresh logging
        if self._reconnect_attempts > 50:  # Reset after many attempts to refresh logging
            self._reconnect_attempts = 10
            
        # Use longer delays if reboot detected
        if self._reboot_detected:
            delay = min(
                5.0 * (2 ** min(self._reconnect_attempts - 1, 3)),  # Start at 5s for reboots
                self._max_reconnect_delay
            ) + random.uniform(0, 2)
        else:
            delay = min(
                self._base_reconnect_delay * (2 ** min(self._reconnect_attempts - 1, 6)),  # Cap exponential growth
                self._max_reconnect_delay
            ) + random.uniform(0, 1)

        # Only log reconnection attempts periodically to avoid spam
        if current_time - self._last_reconnect_log > self._reconnect_log_interval:
            logging.info(
                f"Attempting reconnection #{self._reconnect_attempts} to {self.host}:{self.port} in {delay:.2f}s"
            )
            self._last_reconnect_log = current_time
        else:
            logging.debug(
                f"Reconnection attempt #{self._reconnect_attempts} to {self.host}:{self.port} in {delay:.2f}s"
            )
            
        await asyncio.sleep(delay)

        try:
            await self.close()  # Ensure clean state
            await self.open()
            
            # Log successful reconnection appropriately based on context
            if self._reboot_detected:
                logging.info(
                    f"Successfully reconnected to {self.host}:{self.port} after reboot "
                    f"(took {self._reconnect_attempts} attempts)"
                )
                self._reboot_detected = False  # Clear reboot flag
            else:
                logging.info(
                    f"Successfully reconnected to {self.host}:{self.port} "
                    f"(after {self._reconnect_attempts} attempts)"
                )
            
            self._last_reconnect_log = 0.0  # Reset log throttling on success
            self._reconnect_attempts = 0  # Reset attempts counter
            return True
        except Exception as e:
            # Only log detailed error every few attempts to avoid spam
            if self._reconnect_attempts <= 2 or self._reconnect_attempts % 10 == 0:
                logging.info(
                    f"Reconnection attempt #{self._reconnect_attempts} failed: {e}"
                )
            else:
                logging.debug(
                    f"Reconnection attempt #{self._reconnect_attempts} failed: {e}"
                )
            return False

    async def write(self, data: str):
        """Write data to Seestar with automatic reconnection on connection reset."""
        try:
            if not self.is_connected():
                raise ConnectionError("Not connected to Seestar")

            logging.trace(f"Writing to {self}: {data}")
            data += "\r\n"
            self.writer.write(data.encode())
            await asyncio.wait_for(self.writer.drain(), timeout=self.write_timeout)
            self.written_messages += 1
        except Exception as e:
            if self._is_connection_reset_error(e):
                # Reduce log verbosity for expected reboot scenarios
                if "[Errno 54]" in str(e) or "Connection reset by peer" in str(e):
                    logging.debug(
                        f"Connection to {self.host}:{self.port} lost during write (telescope may be rebooting)"
                    )
                else:
                    logging.info(
                        f"Connection reset detected while writing to {self.host}:{self.port}"
                    )
                await self.close()

                # Attempt reconnection
                if await self._reconnect_with_backoff():
                    logging.debug("Reconnection successful, retrying write operation")
                    # Retry the write operation once after reconnection
                    logging.trace(f"Retrying write to {self}: {data.strip()}")
                    self.writer.write(data.encode())
                    await asyncio.wait_for(
                        self.writer.drain(), timeout=self.write_timeout
                    )
                    self.written_messages += 1
                else:
                    logging.error(
                        "Failed to reconnect after connection reset during write"
                    )
                    raise ConnectionError("Failed to reconnect after connection reset")
            else:
                logging.error(f"Unexpected error while writing to {self}: {e}")
                await self.close()
                raise

    async def read(self) -> str | None:
        """Read data from Seestar with automatic reconnection on connection reset."""
        try:
            if not self.is_connected():
                return None

            data = await asyncio.wait_for(
                self.reader.readuntil(), timeout=self.read_timeout
            )
            self.read_messages += 1
            return data.decode().strip()
        except Exception as e:
            if self._is_connection_reset_error(e):
                # Reduce log verbosity for expected reboot scenarios
                if "[Errno 54]" in str(e) or "Connection reset by peer" in str(e):
                    logging.info(
                        f"Connection to {self.host}:{self.port} lost during read (telescope may be rebooting)"
                    )
                else:
                    logging.info(
                        f"Connection reset detected while reading from {self.host}:{self.port}"
                    )
                await self.close()

                # Attempt reconnection
                if await self._reconnect_with_backoff():
                    logging.debug("Reconnection successful after read failure")
                    # Don't retry the read here, let the caller handle it
                    return None
                else:
                    logging.debug("Failed to reconnect after connection reset")
                    return None
            else:
                logging.error(f"Unexpected error while reading from {self}: {e}")
                await self.close()
                return None

    async def read_exactly(self, n: int) -> bytes | None:
        """Read exactly N bytes from Seestar with automatic reconnection on connection reset."""
        try:
            if not self.is_connected():
                return None

            data = await asyncio.wait_for(
                self.reader.readexactly(n), timeout=self.read_timeout
            )
            self.read_messages += 1
            return data
        except Exception as e:
            if self._is_connection_reset_error(e):
                logging.debug(
                    f"Connection reset detected while reading exactly {n} bytes from {self.host}:{self.port}"
                )
                await self.close()

                # Attempt reconnection
                if await self._reconnect_with_backoff():
                    logging.debug(
                        "Reconnection successful after read_exactly failure"
                    )
                    # Don't retry the read here, let the caller handle it
                    return None
                else:
                    logging.debug("Failed to reconnect after connection reset")
                    return None
            else:
                logging.error(
                    f"Unexpected error while reading exactly {n} bytes from {self}: {e}"
                )
                await self.close()
                return None



"""Async utilities."""

import asyncio

from pydantic import BaseModel
from typing import Any, Awaitable, Callable, Coroutine, Optional


class ResettableDelay(BaseModel):
    """Resettable delay."""

    delay: float
    reset_event: asyncio.Event


# todo : note sure what the above does, but it auto completed ;)
# class AsyncContextManager:
#     """Async context manager."""
#     def __init__(self, func: Callable[..., Awaitable[Any]]) -> None:
#         self._func = func
#         self._instance: Optional[Any] = None
#         self._lock = asyncio.Lock()
#         # pylint: disable=no-self-use
#         self._args: tuple[Any, ...] = ()
#         self._kwargs: dict[str, Any] = {}
#         self._coro: Optional[Coroutine[Any, Any, Any]] = None
#         self._result: Optional[Any] = None
#         self._exception: Optional[Exception] = None
#         self._finished = False
#         self._finished_lock = asyncio.Lock()
#         self._finished_event = asyncio.Event()
#         self._finished_event.set()
#
#     def __call__(self, *args: Any, **kwargs: Any) -> 'AsyncContextManager':
#         self._args = args
#         self._kwargs = kwargs
#         return self
#
#     async def __aenter__(self) -> Any:
#         async with self._lock:
#             if self._instance is None:
#                 self._instance = await self._func(*self._args, **self._kwargs)
#         return self._instance
#
#     async def __aexit__(self, *exc: Any) -> None:
#         async with self._finished_lock:
#             if self._finished:
#                 return
#             self._finished = True
#             self._exception = exc[1] if exc else None
#             self._coro = asyncio.create_task(self._instance.__aexit__(*exc))
#
#



"""Base backend class.

All protocol backends must implement this abstract base class to provide
a unified interface for device discovery and instantiation.
"""

from abc import ABC, abstractmethod
from typing import Optional

from pydantic import BaseModel, Field

from scopinator.v2.core.devices import Camera, FilterWheel, Focuser, Mount
from scopinator.v2.core.events import UnifiedEventBus


class BackendInfo(BaseModel):
    """Information about a backend."""

    name: str = Field(..., description="Backend name")
    version: str = Field(default="1.0.0", description="Backend version")
    description: str = Field(default="", description="Backend description")
    protocol: str = Field(default="unknown", description="Protocol type")


class Backend(ABC):
    """Abstract base class for all backends.

    A backend represents a connection to a device control system
    (e.g., Seestar telescope, ASCOM Alpaca server, INDI server).

    Implementations must provide methods for:
    - Connecting to the backend
    - Discovering available devices
    - Creating device instances (Mount, Camera, etc.)

    Example:
        backend = SeestarBackend(host="192.168.1.100")
        await backend.connect()

        devices = await backend.discover_devices()
        # {'mount': ['seestar_mount'], 'camera': ['seestar_camera']}

        mount = await backend.get_mount('seestar_mount')
        await mount.connect()
    """

    def __init__(self, event_bus: Optional[UnifiedEventBus] = None) -> None:
        """Initialize backend.

        Args:
            event_bus: Optional shared event bus. If not provided, a new
                      one will be created.
        """
        self._event_bus = event_bus or UnifiedEventBus()
        self._connected = False

    @property
    def event_bus(self) -> UnifiedEventBus:
        """Get the event bus for this backend."""
        return self._event_bus

    @event_bus.setter
    def event_bus(self, value: UnifiedEventBus) -> None:
        """Set the event bus (used by DeviceManager)."""
        self._event_bus = value

    @property
    def is_connected(self) -> bool:
        """Check if backend is connected."""
        return self._connected

    @abstractmethod
    async def connect(self) -> None:
        """Connect to the backend.

        For Seestar: Establish TCP connection
        For Alpaca: Initialize HTTP session
        For INDI: Connect to INDI server

        Raises:
            ConnectionError: If connection fails
        """
        pass

    @abstractmethod
    async def disconnect(self) -> None:
        """Disconnect from the backend."""
        pass

    @abstractmethod
    async def discover_devices(self) -> dict[str, list[str]]:
        """Discover available devices.

        Returns:
            Dict mapping device type to list of device IDs.
            Device types: "mount", "camera", "focuser", "filterwheel"

            Example:
                {
                    "mount": ["Celestron CGX"],
                    "camera": ["ZWO ASI294MC Pro"],
                    "focuser": ["ZWO EAF"]
                }
        """
        pass

    @abstractmethod
    async def get_mount(self, device_id: str) -> Mount:
        """Get a mount device by ID.

        Args:
            device_id: Device identifier from discover_devices()

        Returns:
            Mount instance (not yet connected)

        Raises:
            DeviceError: If device not found
        """
        pass

    @abstractmethod
    async def get_camera(self, device_id: str) -> Camera:
        """Get a camera device by ID.

        Args:
            device_id: Device identifier from discover_devices()

        Returns:
            Camera instance (not yet connected)

        Raises:
            DeviceError: If device not found
        """
        pass

    async def get_focuser(self, device_id: str) -> Focuser:
        """Get a focuser device by ID.

        Override if backend supports focusers.

        Args:
            device_id: Device identifier from discover_devices()

        Returns:
            Focuser instance (not yet connected)

        Raises:
            NotSupportedError: If backend doesn't support focusers
        """
        from scopinator.v2.core.exceptions import NotSupportedError

        raise NotSupportedError(f"{self.__class__.__name__} does not support focusers")

    async def get_filterwheel(self, device_id: str) -> FilterWheel:
        """Get a filter wheel device by ID.

        Override if backend supports filter wheels.

        Args:
            device_id: Device identifier from discover_devices()

        Returns:
            FilterWheel instance (not yet connected)

        Raises:
            NotSupportedError: If backend doesn't support filter wheels
        """
        from scopinator.v2.core.exceptions import NotSupportedError

        raise NotSupportedError(
            f"{self.__class__.__name__} does not support filter wheels"
        )

    @abstractmethod
    def get_info(self) -> BackendInfo:
        """Get backend information.

        Returns:
            BackendInfo with name, version, description
        """
        pass

    async def __aenter__(self) -> "Backend":
        """Async context manager entry."""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> bool:
        """Async context manager exit."""
        await self.disconnect()
        return False



"""Interactive mode with autocomplete and intellisense for scopinator CLI."""

import readline
import rlcompleter
import click
from typing import List, Optional, Dict, Any
from scopinator.util.logging_config import get_logger
logger = get_logger(__name__)


class ScopinatorCompleter:
    """Custom completer for scopinator interactive mode."""
    
    def __init__(self, cli_group):
        self.cli_group = cli_group
        self.commands = {}
        self.current_command = None
        self.parse_commands()
        
    def parse_commands(self):
        """Parse all available commands and their parameters."""
        for name, cmd in self.cli_group.commands.items():
            if name != 'interactive':  # Skip the interactive command itself
                self.commands[name] = {
                    'obj': cmd,
                    'params': [],
                    'options': {},
                    'help': cmd.help or cmd.short_help or ""
                }
                
                # Parse parameters and options
                for param in cmd.params:
                    if isinstance(param, click.Option):
                        opt_names = param.opts + param.secondary_opts
                        for opt_name in opt_names:
                            self.commands[name]['options'][opt_name] = {
                                'type': param.type,
                                'help': param.help or "",
                                'required': param.required,
                                'default': param.default,
                                'multiple': param.multiple,
                                'is_flag': param.is_flag
                            }
                    elif isinstance(param, click.Argument):
                        self.commands[name]['params'].append({
                            'name': param.name,
                            'type': param.type,
                            'required': param.required,
                            'nargs': param.nargs
                        })
    
    def get_command_hints(self, command: str) -> str:
        """Get parameter hints for a command."""
        if command not in self.commands:
            return ""
        
        cmd_info = self.commands[command]
        hints = []
        
        # Add required arguments
        for param in cmd_info['params']:
            param_str = f"<{param['name']}>"
            if not param.get('required', True):
                param_str = f"[{param['name']}]"
            hints.append(param_str)
        
        # Add key options
        important_opts = []
        for opt_name, opt_info in cmd_info['options'].items():
            if len(opt_name) <= 3:  # Short options
                if opt_info.get('required'):
                    important_opts.append(f"{opt_name} <value>")
                elif opt_info.get('is_flag'):
                    important_opts.append(f"[{opt_name}]")
        
        if important_opts:
            hints.extend(important_opts[:3])  # Show max 3 options
        
        return " ".join(hints) if hints else ""
    
    def complete(self, text: str, state: int) -> Optional[str]:
        """Readline completer function."""
        # Get the current line buffer
        buffer = readline.get_line_buffer()
        line = buffer.lstrip()
        
        # If we're at the beginning, complete command names
        if not line or line == text:
            matches = [cmd + ' ' for cmd in self.commands.keys() 
                      if cmd.startswith(text)]
            # Add special commands
            special = ['help', 'exit', 'quit', 'clear']
            matches.extend([s + ' ' for s in special if s.startswith(text)])
        else:
            # Parse what command we're working with
            parts = line.split()
            if parts:
                cmd_name = parts[0]
                
                # If this is a known command, provide option completion
                if cmd_name in self.commands:
                    matches = self.complete_command_options(cmd_name, text, parts)
                else:
                    matches = []
            else:
                matches = []
        
        # Return the state-th match
        if state < len(matches):
            return matches[state]
        return None
    
    def complete_command_options(self, cmd_name: str, text: str, parts: List[str]) -> List[str]:
        """Complete options for a specific command."""
        cmd_info = self.commands[cmd_name]
        matches = []
        
        # Get all option names that haven't been used yet
        used_options = set()
        for part in parts[1:]:
            if part.startswith('-'):
                used_options.add(part.split('=')[0])
        
        # Complete option names
        for opt_name in cmd_info['options'].keys():
            if opt_name not in used_options and opt_name.startswith(text):
                if cmd_info['options'][opt_name].get('is_flag'):
                    matches.append(opt_name + ' ')
                else:
                    matches.append(opt_name + '=')
        
        # If completing after an option that needs a value
        if len(parts) >= 2 and parts[-2].startswith('-'):
            opt_name = parts[-2].split('=')[0]
            if opt_name in cmd_info['options']:
                opt_info = cmd_info['options'][opt_name]
                # Provide type-specific suggestions
                if opt_info['type'] == click.INT:
                    matches = ['4700 ', '4800 ', '4900 ']  # Port suggestions
                elif isinstance(opt_info['type'], click.Choice):
                    matches = [c + ' ' for c in opt_info['type'].choices 
                              if c.startswith(text)]
        
        return matches


class InteractiveMode:
    """Enhanced interactive mode with autocomplete."""
    
    def __init__(self, cli_group, ctx):
        self.cli_group = cli_group
        self.ctx = ctx
        self.completer = ScopinatorCompleter(cli_group)
        self.setup_readline()
        
    def setup_readline(self):
        """Configure readline for autocomplete and history."""
        # Set up tab completion
        readline.set_completer(self.completer.complete)
        readline.parse_and_bind('tab: complete')
        
        # Set up history
        import os
        histfile = os.path.expanduser('~/.scopinator_history')
        try:
            readline.read_history_file(histfile)
            readline.set_history_length(1000)
        except FileNotFoundError:
            pass
        
        # Save history on exit
        import atexit
        atexit.register(readline.write_history_file, histfile)
        
        # Configure completion display
        readline.set_completer_delims(' \t\n=')
        readline.set_completion_display_matches_hook(self.display_matches)
    
    def display_matches(self, substitution: str, matches: List[str], longest_match_length: int):
        """Custom display for completion matches."""
        click.echo()
        if len(matches) > 1:
            # Group and display matches
            click.echo("Available completions:")
            for match in sorted(matches):
                # Clean up the match display
                display_match = match.rstrip()
                
                # Check if this is a command and add help text
                cmd_name = display_match.split()[0] if display_match else ""
                if cmd_name in self.completer.commands:
                    help_text = self.completer.commands[cmd_name]['help']
                    if help_text:
                        click.echo(f"  {display_match:<20} - {help_text[:50]}")
                    else:
                        click.echo(f"  {display_match}")
                else:
                    click.echo(f"  {display_match}")
        
        # Redisplay the prompt and current input
        click.echo()
        current_input = readline.get_line_buffer()
        click.echo(f"scopinator> {current_input}", nl=False)
    
    def show_command_help(self, cmd_name: str):
        """Show detailed help for a specific command."""
        if cmd_name in self.completer.commands:
            cmd_info = self.completer.commands[cmd_name]
            cmd = cmd_info['obj']
            
            click.echo(f"\n{click.style(cmd_name, bold=True)} - {cmd_info['help']}")
            
            # Show usage
            usage_parts = [cmd_name]
            hints = self.completer.get_command_hints(cmd_name)
            if hints:
                usage_parts.append(hints)
            click.echo(f"Usage: {' '.join(usage_parts)}")
            
            # Show parameters
            if cmd_info['params']:
                click.echo("\nArguments:")
                for param in cmd_info['params']:
                    param_line = f"  {param['name']}"
                    if param['type']:
                        param_line += f" ({param['type'].name})"
                    if not param.get('required', True):
                        param_line += " [optional]"
                    click.echo(param_line)
            
            # Show options
            if cmd_info['options']:
                click.echo("\nOptions:")
                for opt_name, opt_info in cmd_info['options'].items():
                    if opt_name.startswith('--'):
                        opt_line = f"  {opt_name}"
                        if not opt_info.get('is_flag'):
                            opt_line += f" <{opt_info['type'].name if hasattr(opt_info['type'], 'name') else 'value'}>"
                        if opt_info['help']:
                            opt_line += f" - {opt_info['help']}"
                        click.echo(opt_line)
            click.echo()
        else:
            click.echo(f"Unknown command: {cmd_name}")
    
    def run(self):
        """Run the interactive mode loop."""
        click.echo("Welcome to Scopinator Interactive Mode!")
        click.echo("Type 'help' for commands, Tab for autocomplete, 'exit' to quit.\n")
        
        while True:
            try:
                # Use readline-enabled input
                command = input(click.style("scopinator> ", fg='green', bold=True))
                
                if not command:
                    continue
                
                command = command.strip()
                
                # Handle special commands
                if command.lower() in ['exit', 'quit', 'q']:
                    click.echo("Goodbye!")
                    break
                elif command.lower() == 'clear':
                    click.clear()
                    continue
                elif command.lower() == 'help':
                    self.show_general_help()
                    continue
                elif command.lower().startswith('help '):
                    # Help for specific command
                    cmd_name = command[5:].strip()
                    self.show_command_help(cmd_name)
                    continue
                
                # Parse and execute command
                self.execute_command(command)
                    
            except (KeyboardInterrupt, EOFError):
                click.echo("\nGoodbye!")
                break
            except Exception as e:
                logger.error(f"Interactive mode error: {e}")
                click.echo(f"Error: {e}")
    
    def show_general_help(self):
        """Show general help information."""
        click.echo("\n" + click.style("Available Commands:", bold=True))
        
        # Group commands by category
        categories = {
            'Discovery': ['discover'],
            'Connection': ['connect', 'status'],
            'Control': ['park', 'goto'],
            'Imaging': ['stream'],
            'System': ['version', 'help', 'clear', 'exit']
        }
        
        for category, cmds in categories.items():
            click.echo(f"\n{click.style(category, fg='cyan')}:")
            for cmd_name in cmds:
                if cmd_name in self.completer.commands:
                    cmd_info = self.completer.commands[cmd_name]
                    hints = self.completer.get_command_hints(cmd_name)
                    usage = f"{cmd_name} {hints}".strip()
                    click.echo(f"  {usage:<30} - {cmd_info['help'][:40]}")
                elif cmd_name in ['help', 'clear', 'exit']:
                    helps = {
                        'help': 'Show this help or help for a command',
                        'clear': 'Clear the screen',
                        'exit': 'Exit interactive mode'
                    }
                    click.echo(f"  {cmd_name:<30} - {helps[cmd_name]}")
        
        click.echo(f"\n{click.style('Tips:', bold=True)}")
        click.echo("  • Use Tab for command and option completion")
        click.echo("  • Use arrow keys to navigate command history")  
        click.echo("  • Type 'help <command>' for detailed command help")
        click.echo()
    
    def execute_command(self, command: str):
        """Execute a command in the CLI."""
        args = command.split()
        if not args:
            return
            
        cmd_name = args[0]
        cmd_args = args[1:]
        
        # Find and invoke the command
        cmd = self.cli_group.commands.get(cmd_name)
        if cmd and cmd_name != 'interactive':
            try:
                # Create a new context for the command
                sub_ctx = click.Context(cmd, parent=self.ctx, info_name=cmd_name)
                sub_ctx.obj = self.ctx.obj  # Pass along the context object
                
                # Parse the arguments for the command
                parser = cmd.make_parser(sub_ctx)
                opts, args, order = parser.parse_args(args=cmd_args)
                
                # Invoke the command with parsed arguments
                sub_ctx.params = opts
                cmd.invoke(sub_ctx, **opts)
            except click.ClickException as e:
                e.show()
            except Exception as e:
                click.echo(f"Error: {e}")
        else:
            if cmd_name != 'interactive':
                click.echo(f"Unknown command: {cmd_name}. Type 'help' for available commands.")
                # Suggest similar commands
                similar = [c for c in self.completer.commands.keys() 
                          if c.startswith(cmd_name[:2])]
                if similar:
                    click.echo(f"Did you mean: {', '.join(similar[:3])}?")


"""Seestar backend - adapts existing SeestarClient to V2 interface.

This backend wraps the existing Seestar protocol implementation to
provide the V2 abstraction layer, enabling multi-protocol support
without breaking existing code.
"""

from typing import Optional

from scopinator.seestar.client import SeestarClient
from scopinator.seestar.imaging_client import SeestarImagingClient
from scopinator.v2.backends.base import Backend, BackendInfo
from scopinator.v2.backends.seestar.camera import SeestarCamera
from scopinator.v2.backends.seestar.event_translator import SeestarEventTranslator
from scopinator.v2.backends.seestar.mount import SeestarMount
from scopinator.v2.core.devices import Camera, Focuser, Mount
from scopinator.v2.core.events import EventType, UnifiedEvent, UnifiedEventBus
from scopinator.v2.core.exceptions import ConnectionError, DeviceError


class SeestarBackend(Backend):
    """Backend for Seestar telescopes.

    This is an adapter that wraps the existing SeestarClient and
    SeestarImagingClient to provide the V2 interface. The underlying
    Seestar protocol implementation is preserved and reused.

    Example:
        backend = SeestarBackend(host="192.168.1.100")
        await backend.connect()

        mount = await backend.get_mount("seestar_mount")
        await mount.slew_to_coordinates(Coordinates(ra=83.63, dec=22.01))

        camera = await backend.get_camera("seestar_camera")
        await camera.start_exposure(ExposureSettings(duration_seconds=10.0))
    """

    def __init__(
        self,
        host: str,
        port: int = 4700,
        imaging_port: int = 4800,
        event_bus: Optional[UnifiedEventBus] = None,
    ) -> None:
        """Initialize Seestar backend.

        Args:
            host: Seestar telescope IP address
            port: Control port (default 4700)
            imaging_port: Imaging port (default 4800)
            event_bus: Optional shared event bus
        """
        super().__init__(event_bus)
        self._host = host
        self._port = port
        self._imaging_port = imaging_port

        # The underlying Seestar clients
        self._client: Optional[SeestarClient] = None
        self._imaging_client: Optional[SeestarImagingClient] = None

        # Event translator bridges Seestar events to unified events
        self._event_translator: Optional[SeestarEventTranslator] = None

        # Cached device instances
        self._mount: Optional[SeestarMount] = None
        self._camera: Optional[SeestarCamera] = None

    @property
    def host(self) -> str:
        """Get the Seestar host address."""
        return self._host

    @property
    def port(self) -> int:
        """Get the control port."""
        return self._port

    @property
    def client(self) -> Optional[SeestarClient]:
        """Get the underlying SeestarClient (for advanced use)."""
        return self._client

    @property
    def imaging_client(self) -> Optional[SeestarImagingClient]:
        """Get the underlying SeestarImagingClient (for advanced use)."""
        return self._imaging_client

    async def connect(self) -> None:
        """Connect to Seestar telescope."""
        try:
            # Create the translator first so it can receive events
            self._event_translator = SeestarEventTranslator(self._event_bus)

            # Create clients with the translator's event bus
            self._client = SeestarClient(
                host=self._host,
                port=self._port,
                event_bus=self._event_translator.seestar_event_bus,
            )

            self._imaging_client = SeestarImagingClient(
                host=self._host,
                port=self._imaging_port,
                event_bus=self._event_translator.seestar_event_bus,
            )

            # Connect both clients
            await self._client.connect()
            await self._imaging_client.connect()

            self._connected = True

            # Emit connected event
            event = UnifiedEvent(
                event_type=EventType.CONNECTED,
                source_device="seestar",
                source_backend="seestar",
                data={"host": self._host, "port": self._port},
            )
            self._event_bus.emit_nowait(event)

        except Exception as e:
            self._connected = False
            raise ConnectionError(f"Failed to connect to Seestar at {self._host}: {e}")

    async def disconnect(self) -> None:
        """Disconnect from Seestar telescope."""
        try:
            if self._client:
                await self._client.disconnect()
            if self._imaging_client:
                await self._imaging_client.disconnect()
        finally:
            self._connected = False
            self._client = None
            self._imaging_client = None
            self._mount = None
            self._camera = None

            # Emit disconnected event
            event = UnifiedEvent(
                event_type=EventType.DISCONNECTED,
                source_device="seestar",
                source_backend="seestar",
            )
            self._event_bus.emit_nowait(event)

    async def discover_devices(self) -> dict[str, list[str]]:
        """Discover available devices.

        Seestar is a combined mount+camera device, so discovery always
        returns the same set of virtual devices.

        Returns:
            Dict with mount, camera, and focuser device IDs
        """
        if not self._connected:
            return {}

        return {
            "mount": ["seestar_mount"],
            "camera": ["seestar_camera"],
            "focuser": ["seestar_focuser"],  # Seestar has built-in focuser
        }

    async def get_mount(self, device_id: str) -> Mount:
        """Get Seestar mount interface.

        Args:
            device_id: Device ID (typically "seestar_mount")

        Returns:
            SeestarMount instance
        """
        if not self._client:
            raise DeviceError("Not connected to Seestar")

        if self._mount is None:
            self._mount = SeestarMount(
                client=self._client,
                event_bus=self._event_bus,
            )
            await self._mount.connect()

        return self._mount

    async def get_camera(self, device_id: str) -> Camera:
        """Get Seestar camera interface.

        Args:
            device_id: Device ID (typically "seestar_camera")

        Returns:
            SeestarCamera instance
        """
        if not self._client:
            raise DeviceError("Not connected to Seestar")

        if self._camera is None:
            self._camera = SeestarCamera(
                client=self._client,
                imaging_client=self._imaging_client,
                event_bus=self._event_bus,
            )
            await self._camera.connect()

        return self._camera

    async def get_focuser(self, device_id: str) -> Focuser:
        """Get Seestar focuser interface.

        Note: Seestar focuser support is not yet fully implemented
        in the V2 layer. The underlying client supports focuser
        commands, but the V2 Focuser adapter needs to be created.

        Args:
            device_id: Device ID (typically "seestar_focuser")

        Raises:
            NotSupportedError: Focuser V2 adapter not yet implemented
        """
        from scopinator.v2.core.exceptions import NotSupportedError

        raise NotSupportedError("Seestar focuser not yet implemented in V2")

    def get_info(self) -> BackendInfo:
        """Get backend information."""
        return BackendInfo(
            name="Seestar",
            version="1.0.0",
            description="ZWO Seestar smart telescope backend",
            protocol="seestar",
        )

    def __repr__(self) -> str:
        return f"SeestarBackend(host={self._host}, port={self._port}, connected={self._connected})"



"""Enhanced response models for parsing telescope message data."""

from typing import Any, Literal, Optional, Union
from pydantic import BaseModel, Field, field_validator
from datetime import datetime

from .common import CommandResponse
from .simple import (
    GetTimeResponse,
    GetCameraInfoResponse,
    GetCameraStateResponse,
    GetDiskVolumeResponse,
    GetDeviceStateResponse,
)


class MessageParsingError(BaseModel):
    """Error information when message parsing fails."""

    error_type: str
    message: str
    raw_data: str
    timestamp: str


class ParsedMessage(BaseModel):
    """Base class for parsed telescope messages."""

    raw_message: str
    timestamp: str
    message_type: Literal["command", "response", "event", "unknown"]
    parse_success: bool = True
    parse_error: Optional[MessageParsingError] = None


class ParsedCommand(ParsedMessage):
    """Parsed command message sent to telescope."""

    message_type: Literal["command"] = "command"
    command_id: Optional[int] = None
    method: str
    params: Optional[dict[str, Any]] = None


class ParsedResponse(ParsedMessage):
    """Parsed response message received from telescope."""

    message_type: Literal["response"] = "response"
    response_id: int
    method: str
    jsonrpc: str = "2.0"
    code: int
    result: Optional[Union[dict[str, Any], list[Any], int, float, str]] = None
    error: Optional[dict[str, Any]] = None
    timestamp_telescope: Optional[str] = None  # Telescope's timestamp

    @field_validator("timestamp_telescope", mode="before")
    def parse_telescope_timestamp(cls, v):
        """Parse telescope timestamp if present."""
        if isinstance(v, str):
            try:
                # Validate it's a proper ISO format
                datetime.fromisoformat(v.replace("Z", "+00:00"))
                return v
            except ValueError:
                return None
        return v


class ParsedEvent(ParsedMessage):
    """Parsed event message received from telescope."""

    message_type: Literal["event"] = "event"
    event_type: str
    event_data: dict[str, Any]
    timestamp_telescope: Optional[str] = None


class UnknownMessage(ParsedMessage):
    """Message that couldn't be categorized."""

    message_type: Literal["unknown"] = "unknown"
    attempted_parse_as: list[str] = []


# Specific enhanced response models with better typing


class EnhancedCommandResponse(CommandResponse):
    """Enhanced command response with better result typing."""

    parsed_result: Optional[BaseModel] = None
    result_type: Optional[str] = None

    def parse_result_as(self, model_class: type[BaseModel]) -> Optional[BaseModel]:
        """Parse the result field as a specific model type."""
        if self.result is None:
            return None

        try:
            if isinstance(self.result, dict):
                parsed = model_class(**self.result)
                self.parsed_result = parsed
                self.result_type = model_class.__name__
                return parsed
            elif isinstance(self.result, (list, tuple)) and hasattr(
                model_class, "__annotations__"
            ):
                # Handle tuple results by converting to dict with field names
                field_names = list(model_class.__annotations__.keys())
                if len(self.result) <= len(field_names):
                    result_dict = {
                        field_names[i]: self.result[i] for i in range(len(self.result))
                    }
                    parsed = model_class(**result_dict)
                    self.parsed_result = parsed
                    self.result_type = model_class.__name__
                    return parsed
            return None
        except Exception:
            return None


class TimeResponse(EnhancedCommandResponse):
    """Response from pi_get_time command."""

    def get_parsed_time(self) -> Optional[GetTimeResponse]:
        """Get parsed time response."""
        return self.parse_result_as(GetTimeResponse)


class DeviceStateResponse(EnhancedCommandResponse):
    """Response from get_device_state command."""

    def get_parsed_device_state(self) -> Optional[GetDeviceStateResponse]:
        """Get parsed device state response."""
        return self.parse_result_as(GetDeviceStateResponse)


class DiskVolumeResponse(EnhancedCommandResponse):
    """Response from get_disk_volume command."""

    def get_parsed_disk_volume(self) -> Optional[GetDiskVolumeResponse]:
        """Get parsed disk volume response."""
        return self.parse_result_as(GetDiskVolumeResponse)


class CameraInfoResponse(EnhancedCommandResponse):
    """Response from get_camera_info command."""

    def get_parsed_camera_info(self) -> Optional[GetCameraInfoResponse]:
        """Get parsed camera info response."""
        return self.parse_result_as(GetCameraInfoResponse)


class CoordinateResponse(EnhancedCommandResponse):
    """Response from coordinate commands (RA/Dec, etc)."""

    def get_coordinates(self) -> Optional[dict[str, float]]:
        """Get coordinates as a dictionary."""
        if isinstance(self.result, dict):
            return self.result
        elif isinstance(self.result, (list, tuple)) and len(self.result) == 2:
            return {"ra": self.result[0], "dec": self.result[1]}
        return None


class FocuserPositionResponse(EnhancedCommandResponse):
    """Response from get_focuser_position command."""

    def get_position(self) -> Optional[int]:
        """Get focuser position as integer."""
        if isinstance(self.result, (int, float)):
            return int(self.result)
        elif isinstance(self.result, dict) and "position" in self.result:
            return self.result["position"]
        return None


class ViewStateResponse(EnhancedCommandResponse):
    """Response from get_view_state command."""

    def get_view_data(self) -> Optional[dict[str, Any]]:
        """Get view state data."""
        if isinstance(self.result, dict) and "View" in self.result:
            return self.result["View"]
        return self.result if isinstance(self.result, dict) else None


# Message parsing utilities


class TelescopeMessageParser:
    """Utility class for parsing telescope messages."""

    @staticmethod
    def parse_message(raw_message: str, timestamp: str) -> ParsedMessage:
        """Parse a raw telescope message into a structured format."""
        import json

        try:
            data = json.loads(raw_message)
        except json.JSONDecodeError as e:
            return UnknownMessage(
                raw_message=raw_message,
                timestamp=timestamp,
                parse_success=False,
                parse_error=MessageParsingError(
                    error_type="JSON_DECODE_ERROR",
                    message=str(e),
                    raw_data=raw_message,
                    timestamp=timestamp,
                ),
            )

        # Determine message type and parse accordingly
        if "Event" in data:
            return TelescopeMessageParser._parse_event(data, raw_message, timestamp)
        elif "jsonrpc" in data:
            return TelescopeMessageParser._parse_response(data, raw_message, timestamp)
        elif "method" in data:
            return TelescopeMessageParser._parse_command(data, raw_message, timestamp)
        else:
            return UnknownMessage(
                raw_message=raw_message,
                timestamp=timestamp,
                attempted_parse_as=["event", "response", "command"],
            )

    @staticmethod
    def _parse_command(
        data: dict[str, Any], raw_message: str, timestamp: str
    ) -> ParsedCommand:
        """Parse a command message."""
        return ParsedCommand(
            raw_message=raw_message,
            timestamp=timestamp,
            command_id=data.get("id"),
            method=data.get("method", "unknown"),
            params=data.get("params"),
        )

    @staticmethod
    def _parse_response(
        data: dict[str, Any], raw_message: str, timestamp: str
    ) -> ParsedResponse:
        """Parse a response message."""
        return ParsedResponse(
            raw_message=raw_message,
            timestamp=timestamp,
            response_id=data.get("id", -1),
            method=data.get("method", "unknown"),
            jsonrpc=data.get("jsonrpc", "2.0"),
            code=data.get("code", 0),
            result=data.get("result"),
            error=data.get("error"),
            timestamp_telescope=data.get("Timestamp"),
        )

    @staticmethod
    def _parse_event(
        data: dict[str, Any], raw_message: str, timestamp: str
    ) -> ParsedEvent:
        """Parse an event message."""
        return ParsedEvent(
            raw_message=raw_message,
            timestamp=timestamp,
            event_type=data.get("Event", "unknown"),
            event_data=data,
            timestamp_telescope=data.get("Timestamp"),
        )

    @staticmethod
    def create_enhanced_response(
        response_data: dict[str, Any],
    ) -> EnhancedCommandResponse:
        """Create an enhanced command response from raw data."""
        base_response = EnhancedCommandResponse(**response_data)

        # Try to create specialized response based on method
        method = response_data.get("method", "")
        if method == "pi_get_time":
            return TimeResponse(**response_data)
        elif method == "get_device_state":
            return DeviceStateResponse(**response_data)
        elif method == "get_disk_volume":
            return DiskVolumeResponse(**response_data)
        elif method == "get_camera_info":
            return CameraInfoResponse(**response_data)
        elif method in [
            "scope_get_equ_coord",
            "scope_get_ra_dec",
            "scope_get_horiz_coord",
        ]:
            return CoordinateResponse(**response_data)
        elif method == "get_focuser_position":
            return FocuserPositionResponse(**response_data)
        elif method == "get_view_state":
            return ViewStateResponse(**response_data)
        else:
            return base_response


# Message analysis utilities


class MessageAnalytics:
    """Utilities for analyzing telescope message patterns."""

    @staticmethod
    def analyze_message_history(messages: list[dict[str, Any]]) -> dict[str, Any]:
        """Analyze a list of telescope messages for patterns and statistics."""
        if not messages:
            return {"error": "No messages to analyze"}

        stats = {
            "total_messages": len(messages),
            "sent_count": 0,
            "received_count": 0,
            "commands": {},
            "events": {},
            "responses": {},
            "parse_errors": 0,
            "time_range": {},
            "most_common_commands": [],
            "most_common_events": [],
            "response_times": [],
        }

        command_counts = {}
        event_counts = {}
        response_counts = {}
        timestamps = []

        for msg in messages:
            direction = msg.get("direction", "unknown")
            if direction == "sent":
                stats["sent_count"] += 1
            elif direction == "received":
                stats["received_count"] += 1

            # Parse timestamp
            timestamp_str = msg.get("timestamp", "")
            if timestamp_str:
                timestamps.append(timestamp_str)

            # Parse message content
            parsed = TelescopeMessageParser.parse_message(
                msg.get("message", ""), timestamp_str
            )

            if not parsed.parse_success:
                stats["parse_errors"] += 1
                continue

            # Count message types
            if isinstance(parsed, ParsedCommand):
                method = parsed.method
                command_counts[method] = command_counts.get(method, 0) + 1
            elif isinstance(parsed, ParsedEvent):
                event_type = parsed.event_type
                event_counts[event_type] = event_counts.get(event_type, 0) + 1
            elif isinstance(parsed, ParsedResponse):
                method = parsed.method
                response_counts[method] = response_counts.get(method, 0) + 1

        # Calculate most common
        stats["commands"] = command_counts
        stats["events"] = event_counts
        stats["responses"] = response_counts

        stats["most_common_commands"] = sorted(
            command_counts.items(), key=lambda x: x[1], reverse=True
        )[:10]

        stats["most_common_events"] = sorted(
            event_counts.items(), key=lambda x: x[1], reverse=True
        )[:10]

        # Time range analysis
        if timestamps:
            timestamps.sort()
            stats["time_range"] = {
                "earliest": timestamps[0],
                "latest": timestamps[-1],
                "duration_messages": len(timestamps),
            }

        return stats



"""Abstract device interfaces.

This module defines the abstract base classes for all device types.
Backend implementations (Seestar, Alpaca, INDI) must implement these
interfaces to provide a unified API.
"""

from abc import ABC, abstractmethod
from typing import AsyncIterator, Optional

from pydantic import BaseModel, Field

from scopinator.v2.core.capabilities import (
    CameraCapabilities,
    FilterWheelCapabilities,
    FocuserCapabilities,
    MountCapabilities,
)
from scopinator.v2.core.events import UnifiedEventBus
from scopinator.v2.core.types import (
    AltAzCoordinates,
    CameraState,
    Coordinates,
    ExposureSettings,
    FilterPosition,
    FocuserPosition,
    ImageData,
    PierSide,
    SlewState,
    TrackingRate,
)


class DeviceStatus(BaseModel):
    """Base device status."""

    connected: bool = Field(default=False, description="Connection state")
    name: str = Field(default="", description="Device name")
    driver_info: str = Field(default="", description="Driver/backend info")
    driver_version: str = Field(default="", description="Driver version")
    description: str = Field(default="", description="Device description")


class Device(ABC):
    """Base class for all devices.

    All device implementations must inherit from this class and implement
    the abstract methods for connection management and status retrieval.
    """

    def __init__(self, event_bus: Optional[UnifiedEventBus] = None) -> None:
        """Initialize device.

        Args:
            event_bus: Optional event bus for emitting device events
        """
        self._event_bus = event_bus or UnifiedEventBus()
        self._connected = False

    @property
    def event_bus(self) -> UnifiedEventBus:
        """Get the event bus for this device."""
        return self._event_bus

    @property
    def is_connected(self) -> bool:
        """Check if device is connected."""
        return self._connected

    @abstractmethod
    async def connect(self) -> None:
        """Connect to the device.

        Raises:
            ConnectionError: If connection fails
        """
        pass

    @abstractmethod
    async def disconnect(self) -> None:
        """Disconnect from the device."""
        pass

    @abstractmethod
    async def get_status(self) -> DeviceStatus:
        """Get current device status.

        Returns:
            DeviceStatus with current state
        """
        pass


# =============================================================================
# Mount
# =============================================================================


class MountStatus(DeviceStatus):
    """Mount-specific status."""

    coordinates: Optional[Coordinates] = Field(None, description="Current RA/Dec")
    altaz: Optional[AltAzCoordinates] = Field(None, description="Current Alt/Az")
    state: SlewState = Field(default=SlewState.IDLE, description="Current slew state")
    tracking: bool = Field(default=False, description="Tracking enabled")
    tracking_rate: TrackingRate = Field(
        default=TrackingRate.SIDEREAL, description="Tracking rate"
    )
    pier_side: PierSide = Field(default=PierSide.UNKNOWN, description="Pier side (GEM)")
    at_park: bool = Field(default=False, description="Is parked")
    at_home: bool = Field(default=False, description="Is at home position")
    slewing: bool = Field(default=False, description="Is slewing")
    target_coordinates: Optional[Coordinates] = Field(None, description="Target coordinates")


class Mount(Device):
    """Abstract mount interface.

    Provides unified control for telescope mounts across different protocols.
    All coordinate-based methods use degrees for RA (0-360) and Dec (-90 to 90).
    """

    @abstractmethod
    async def get_capabilities(self) -> MountCapabilities:
        """Get mount capabilities.

        Returns:
            MountCapabilities describing what this mount can do
        """
        pass

    @abstractmethod
    async def get_coordinates(self) -> Coordinates:
        """Get current equatorial coordinates.

        Returns:
            Current RA/Dec position

        Raises:
            DeviceError: If coordinates cannot be read
        """
        pass

    async def get_altaz(self) -> Optional[AltAzCoordinates]:
        """Get current altitude-azimuth coordinates.

        Override if supported by the backend.

        Returns:
            Current Alt/Az position, or None if not supported
        """
        return None

    @abstractmethod
    async def slew_to_coordinates(
        self,
        coords: Coordinates,
        *,
        wait: bool = True,
    ) -> None:
        """Slew to equatorial coordinates.

        Args:
            coords: Target coordinates
            wait: If True, wait for slew to complete before returning

        Raises:
            DeviceError: If slew fails
            TimeoutError: If wait=True and slew doesn't complete
        """
        pass

    async def slew_to_altaz(
        self,
        coords: AltAzCoordinates,
        *,
        wait: bool = True,
    ) -> None:
        """Slew to altitude-azimuth coordinates.

        Override if supported by the backend.

        Args:
            coords: Target Alt/Az coordinates
            wait: If True, wait for slew to complete

        Raises:
            NotSupportedError: If Alt/Az slewing not supported
        """
        from scopinator.v2.core.exceptions import NotSupportedError

        raise NotSupportedError("Alt/Az slewing not supported by this mount")

    @abstractmethod
    async def abort_slew(self) -> None:
        """Abort any current slew operation."""
        pass

    @abstractmethod
    async def sync_to_coordinates(self, coords: Coordinates) -> None:
        """Sync mount to given coordinates.

        This tells the mount that it is currently pointing at the given
        coordinates, for alignment correction.

        Args:
            coords: Coordinates to sync to

        Raises:
            DeviceError: If sync fails
        """
        pass

    @abstractmethod
    async def park(self) -> None:
        """Park the mount.

        Raises:
            DeviceError: If parking fails
        """
        pass

    async def unpark(self) -> None:
        """Unpark the mount.

        Override if supported by the backend.

        Raises:
            NotSupportedError: If unparking not supported
        """
        from scopinator.v2.core.exceptions import NotSupportedError

        raise NotSupportedError("Unpark not supported by this mount")

    @abstractmethod
    async def set_tracking(self, enabled: bool) -> None:
        """Enable or disable tracking.

        Args:
            enabled: True to enable tracking, False to disable
        """
        pass

    @abstractmethod
    async def get_tracking(self) -> bool:
        """Get current tracking state.

        Returns:
            True if tracking is enabled
        """
        pass

    async def set_tracking_rate(self, rate: TrackingRate) -> None:
        """Set tracking rate.

        Override if supported by the backend.

        Args:
            rate: Desired tracking rate

        Raises:
            NotSupportedError: If custom tracking rates not supported
        """
        from scopinator.v2.core.exceptions import NotSupportedError

        raise NotSupportedError("Custom tracking rates not supported by this mount")

    async def find_home(self) -> None:
        """Find home position.

        Override if supported by the backend.

        Raises:
            NotSupportedError: If homing not supported
        """
        from scopinator.v2.core.exceptions import NotSupportedError

        raise NotSupportedError("Homing not supported by this mount")

    async def pulse_guide(self, direction: str, duration_ms: int) -> None:
        """Pulse guide in a direction.

        Override if supported by the backend.

        Args:
            direction: One of "north", "south", "east", "west"
            duration_ms: Duration in milliseconds

        Raises:
            NotSupportedError: If pulse guiding not supported
        """
        from scopinator.v2.core.exceptions import NotSupportedError

        raise NotSupportedError("Pulse guiding not supported by this mount")

    async def is_slewing(self) -> bool:
        """Check if mount is currently slewing.

        Returns:
            True if slewing
        """
        status = await self.get_status()
        return status.slewing

    @abstractmethod
    async def get_status(self) -> MountStatus:
        """Get current mount status.

        Returns:
            MountStatus with current state
        """
        pass


# =============================================================================
# Camera
# =============================================================================


class CameraStatus(DeviceStatus):
    """Camera-specific status."""

    state: CameraState = Field(default=CameraState.IDLE, description="Camera state")
    exposure_progress: float = Field(
        default=0.0, ge=0.0, le=1.0, description="Exposure progress (0.0 to 1.0)"
    )
    cooler_on: bool = Field(default=False, description="Cooler enabled")
    cooler_setpoint: Optional[float] = Field(None, description="Target temperature")
    temperature: Optional[float] = Field(None, description="Current sensor temperature")
    cooler_power: Optional[float] = Field(None, description="Cooler power percentage")
    gain: Optional[int] = Field(None, description="Current gain setting")
    offset: Optional[int] = Field(None, description="Current offset setting")
    binning: tuple[int, int] = Field(default=(1, 1), description="Current binning (x, y)")
    last_exposure_duration: Optional[float] = Field(None, description="Last exposure duration")


class Camera(Device):
    """Abstract camera interface.

    Provides unified control for cameras across different protocols.
    Supports both single exposure and continuous/stacking modes.
    """

    @abstractmethod
    async def get_capabilities(self) -> CameraCapabilities:
        """Get camera capabilities.

        Returns:
            CameraCapabilities describing what this camera can do
        """
        pass

    @abstractmethod
    async def start_exposure(self, settings: ExposureSettings) -> None:
        """Start an exposure.

        Args:
            settings: Exposure settings

        Raises:
            DeviceError: If exposure cannot be started
        """
        pass

    @abstractmethod
    async def abort_exposure(self) -> None:
        """Abort current exposure."""
        pass

    async def stop_exposure(self) -> None:
        """Stop exposure early and read out.

        Override if supported by the backend.

        Raises:
            NotSupportedError: If stop not supported (use abort instead)
        """
        from scopinator.v2.core.exceptions import NotSupportedError

        raise NotSupportedError("Stop exposure not supported - use abort")

    @abstractmethod
    async def get_image(self) -> ImageData:
        """Get the last captured image.

        Returns:
            ImageData containing the image

        Raises:
            DeviceError: If no image available
        """
        pass

    @abstractmethod
    async def is_exposing(self) -> bool:
        """Check if camera is currently exposing.

        Returns:
            True if exposing
        """
        pass

    async def is_image_ready(self) -> bool:
        """Check if an image is ready to download.

        Returns:
            True if image ready
        """
        return not await self.is_exposing()

    async def stream_images(
        self, settings: ExposureSettings
    ) -> AsyncIterator[ImageData]:
        """Stream continuous images.

        Override if supported by the backend. Default implementation
        takes single exposures in a loop.

        Args:
            settings: Exposure settings for each frame

        Yields:
            ImageData for each captured frame
        """
        while True:
            await self.start_exposure(settings)
            while await self.is_exposing():
                import asyncio
                await asyncio.sleep(0.1)
            yield await self.get_image()

    async def set_cooler(
        self, enabled: bool, setpoint: Optional[float] = None
    ) -> None:
        """Control cooler.

        Override if supported by the backend.

        Args:
            enabled: True to enable cooler
            setpoint: Target temperature in Celsius

        Raises:
            NotSupportedError: If camera has no cooler
        """
        from scopinator.v2.core.exceptions import NotSupportedError

        raise NotSupportedError("This camera does not have a cooler")

    async def get_temperature(self) -> Optional[float]:
        """Get current sensor temperature.

        Returns:
            Temperature in Celsius, or None if not available
        """
        status = await self.get_status()
        return status.temperature

    async def set_gain(self, gain: int) -> None:
        """Set camera gain.

        Args:
            gain: Gain value

        Raises:
            DeviceError: If gain cannot be set
        """
        pass

    async def set_binning(self, bin_x: int, bin_y: int) -> None:
        """Set camera binning.

        Args:
            bin_x: Horizontal binning
            bin_y: Vertical binning

        Raises:
            DeviceError: If binning cannot be set
        """
        pass

    @abstractmethod
    async def get_status(self) -> CameraStatus:
        """Get current camera status.

        Returns:
            CameraStatus with current state
        """
        pass


# =============================================================================
# Focuser
# =============================================================================


class FocuserStatus(DeviceStatus):
    """Focuser-specific status."""

    position: int = Field(default=0, description="Current position in steps")
    max_position: int = Field(default=0, description="Maximum position")
    is_moving: bool = Field(default=False, description="Is currently moving")
    temperature: Optional[float] = Field(None, description="Temperature in Celsius")
    temp_comp_enabled: bool = Field(default=False, description="Temp compensation enabled")


class Focuser(Device):
    """Abstract focuser interface."""

    @abstractmethod
    async def get_capabilities(self) -> FocuserCapabilities:
        """Get focuser capabilities.

        Returns:
            FocuserCapabilities describing what this focuser can do
        """
        pass

    @abstractmethod
    async def get_position(self) -> FocuserPosition:
        """Get current focuser position.

        Returns:
            Current position information
        """
        pass

    @abstractmethod
    async def move_to(self, position: int, *, wait: bool = True) -> None:
        """Move to absolute position.

        Args:
            position: Target position in steps
            wait: If True, wait for move to complete

        Raises:
            DeviceError: If move fails
            ValueError: If position out of range
        """
        pass

    @abstractmethod
    async def move_relative(self, steps: int, *, wait: bool = True) -> None:
        """Move relative to current position.

        Args:
            steps: Number of steps (positive = out, negative = in)
            wait: If True, wait for move to complete

        Raises:
            DeviceError: If move fails
        """
        pass

    @abstractmethod
    async def halt(self) -> None:
        """Halt any movement."""
        pass

    async def is_moving(self) -> bool:
        """Check if focuser is moving.

        Returns:
            True if moving
        """
        status = await self.get_status()
        return status.is_moving

    async def get_temperature(self) -> Optional[float]:
        """Get focuser temperature.

        Returns:
            Temperature in Celsius, or None if not available
        """
        position = await self.get_position()
        return position.temperature

    @abstractmethod
    async def get_status(self) -> FocuserStatus:
        """Get current focuser status.

        Returns:
            FocuserStatus with current state
        """
        pass


# =============================================================================
# Filter Wheel
# =============================================================================


class FilterWheelStatus(DeviceStatus):
    """Filter wheel-specific status."""

    position: int = Field(default=0, description="Current position (0-indexed)")
    filter_name: Optional[str] = Field(None, description="Current filter name")
    is_moving: bool = Field(default=False, description="Is currently moving")
    num_positions: int = Field(default=0, description="Number of filter slots")


class FilterWheel(Device):
    """Abstract filter wheel interface."""

    @abstractmethod
    async def get_capabilities(self) -> FilterWheelCapabilities:
        """Get filter wheel capabilities.

        Returns:
            FilterWheelCapabilities describing this filter wheel
        """
        pass

    @abstractmethod
    async def get_position(self) -> FilterPosition:
        """Get current filter position.

        Returns:
            Current filter position and name
        """
        pass

    @abstractmethod
    async def set_position(self, position: int, *, wait: bool = True) -> None:
        """Move to filter position.

        Args:
            position: Target position (0-indexed)
            wait: If True, wait for move to complete

        Raises:
            DeviceError: If move fails
            ValueError: If position out of range
        """
        pass

    async def set_filter_by_name(self, name: str, *, wait: bool = True) -> None:
        """Move to filter by name.

        Args:
            name: Filter name
            wait: If True, wait for move to complete

        Raises:
            ValueError: If filter name not found
        """
        names = await self.get_filter_names()
        try:
            position = names.index(name)
        except ValueError:
            raise ValueError(f"Filter '{name}' not found. Available: {names}")
        await self.set_position(position, wait=wait)

    @abstractmethod
    async def get_filter_names(self) -> list[str]:
        """Get list of filter names.

        Returns:
            List of filter names by position
        """
        pass

    async def is_moving(self) -> bool:
        """Check if filter wheel is moving.

        Returns:
            True if moving
        """
        status = await self.get_status()
        return status.is_moving

    @abstractmethod
    async def get_status(self) -> FilterWheelStatus:
        """Get current filter wheel status.

        Returns:
            FilterWheelStatus with current state
        """
        pass
